import matplotlib.pyplot as plt
import numpy as np
import tensorflow as tf

from tf_pwa.config_loader import ConfigLoader
from tf_pwa.data import data_index, data_split
from tf_pwa.histogram import Hist1D


def sum_resolution(amps, weights, size=1):
    amps = tf.reshape(amps * weights, (-1, size))
    amps = tf.reduce_sum(amps, axis=-1).numpy()
    return amps


def main():
    config = ConfigLoader("config.yml")
    config.set_params("final_params.json")
    amp = config.get_amplitude()

    data = config.get_data("data_origin")[0]
    phsp = config.get_data("phsp_plot")[0]
    phsp_re = config.get_data("phsp_plot_re")[0]

    print("data loaded")
    amps = amp(phsp_re)
    pw = amp.partial_weight(phsp_re)

    re_weight = phsp_re["weight"]
    re_size = config.resolution_size
    amps = sum_resolution(amps, re_weight, re_size)
    pw = [sum_resolution(i, re_weight, re_size) for i in pw]

    m_idx = config.get_data_index("mass", "R_BC")
    m_phsp = data_index(phsp, m_idx).numpy()
    m_data = data_index(data, m_idx).numpy()

    m_min, m_max = np.min(m_phsp), np.max(m_phsp)

    scale = m_data.shape[0] / np.sum(amps)

    get_hist = lambda m, w: Hist1D.histogram(
        m, weights=w, range=(m_min, m_max), bins=100
    )

    data_hist = get_hist(m_data, None)
    phsp_hist = get_hist(m_phsp, scale * amps)
    pw_hist = []
    for i in pw:
        pw_hist.append(get_hist(m_phsp, scale * i))

    ax2 = plt.subplot2grid((4, 1), (3, 0), rowspan=1)
    ax = plt.subplot2grid((4, 1), (0, 0), rowspan=3, sharex=ax2)
    data_hist.draw_error(ax, label="data")
    phsp_hist.draw(ax, label="fit")

    for i, j in zip(pw_hist, config.get_decay()):
        i.draw_kde(ax, label=str(j.inner[0]))

    (data_hist - phsp_hist).draw_pull(ax2)
    ax.set_ylim((1, None))
    ax.legend()
    ax.set_yscale("log")
    ax.set_ylabel("Events/{:.1f} MeV".format((m_max - m_min) * 10))
    ax2.set_xlabel("M( R_BC )")
    ax2.set_ylabel("pull")
    ax2.set_xlim((1.3, 1.7))
    ax2.set_ylim((-5, 5))
    plt.setp(ax.get_xticklabels(), visible=False)
    plt.savefig("m_R_BC_fit.png")


if __name__ == "__main__":
    main()



"""
This module implements classes and methods to manage the variables in fitting.
"""
import contextlib
import warnings

import numpy as np
import sympy as sy

from .config import get_config, regist_config
from .params_trans import ParamsTrans
from .tensorflow_wrapper import tf


def combineVM(vm1, vm2, name="", same_list=None):
    """
    This function combines two VarsManager objects into one. (WIP)

    :param name: The name of this combined VarsManager
    :param same_list: To make some variables in the two VarsManager to be the same. E.g. if ``same_list = ["var",["var1","var2"]]``, then "var" in vm1 and vm2 will be the same, and "var1" in vm1 and "var2" in vm2 will be the same.
    """
    if same_list is None:
        same_list = []
    if vm1.name == vm2.name:
        raise Exception(
            "The two VarsManager to be combined have the same name."
        )
    vm = VarsManager(name, vm1.dtype)

    for i in vm1.variables:
        ii = vm1.name + i
        vm.variables[ii] = vm1.variables[i]  # vm.variables
        if vm.variables[ii].trainable:
            vm.trainable_vars.append(ii)  # vm.trainable_vars
        if i in vm1.bnd_dic:
            vm.bnd_dic[ii] = vm1.bnd_dic[i]  # vm.bnd_dic
    for i in vm2.variables:
        ii = vm2.name + i
        vm.variables[ii] = vm2.variables[i]
        if vm.variables[ii].trainable:
            vm.trainable_vars.append(ii)
        if i in vm2.bnd_dic:
            vm.bnd_dic[ii] = vm2.bnd_dic[i]

    for i in vm1.complex_vars:
        ii = vm1.name + i
        vm.complex_vars[ii] = vm1.complex_vars[i]  # vm.complex_vars
    for i in vm2.complex_vars:
        ii = vm2.name + i
        vm.complex_vars[ii] = vm2.complex_vars[i]

    for i in vm1.same_list:
        vm.same_list.append([vm1.name + j for j in i])  # vm.same_list
    for i in vm2.same_list:
        vm.same_list.append([vm2.name + j for j in i])

    for V in vm1.var_head:
        vm.var_head[V] = [vm1.name + i for i in vm1.var_head[V]]  # vm.var_head
        V.name = vm1.name + V.name
        V.vm = vm
    for V in vm2.var_head:
        vm.var_head[V] = [vm2.name + i for i in vm2.var_head[V]]
        V.name = vm2.name + V.name
        V.vm = vm

    for i in same_list:
        if type(i) == str:
            vm.set_same([vm1.name + i, vm2.name + i])
        else:
            vm.set_same([vm1.name + i[0], vm2.name + i[1]])

    del vm1, vm2
    return vm


regist_config("polar", True)


class VarsManager(object):
    """
    This class provides methods to operate the variables in fitting. Every variable is a 1-d **tf.Variable** of
    **dtype** (**tf.float64** by default).

    All variables are stored in a dictionary **self.variables**. The indices of the dictionary are the variables' names,
    so name property in **tf.Variable** does not matter. All methods intended to change the variables are operating
    **self.variables** directly.

    Besides, all trainable variables' names will be stored in a list **self.trainable_vars**.
    """

    def __init__(self, name="", dtype=tf.float64):
        self.name = name
        self.dtype = dtype
        self.polar = get_config("polar")
        self.variables = {}  # {name:tf.Variable,...}
        self.trainable_vars = []  # [name,...]
        self.complex_vars = {}  # {name:polar(bool),...}
        self.same_list = []  # [[name1,name2],...]

        self.bnd_dic = {}  # {name:(a,b),...}

        self.var_head = (
            {}
        )  # {head:[name1,name2],...} It's operated directly by Variable objects

        self.init_val = {}

    def add_real_var(self, name, value=None, range_=None, trainable=True):
        """
        Add a real variable named **name** into **self.variables**. If **value** and **range_** are not provided, the initial value
        is set to be a uniform random number between 0 and 1.

        :param name: The name of the variable, the index of this variable in **self.variables**
        :param value: The initial value.
        :param range_: Length-2 array. It's useless if **value** is given. Otherwise the initial value is set to be a uniform random number between **range_[0]** and **range_[0]**.
        :param trainable: Boolean. If it's **True**, the variable is trainable while fitting.
        """
        if name in self.variables:  # not a new var
            if name in self.trainable_vars:
                self.trainable_vars.remove(name)
            # warnings.warn("Overwrite variable {}!".format(name))

        if value is None:
            if range_ is None:  # random [0,1]
                self.variables[name] = tf.Variable(
                    tf.random.uniform(
                        shape=[], minval=0.0, maxval=1.0, dtype=self.dtype
                    ),
                    trainable=trainable,
                )
            else:  # random [a,b]
                self.variables[name] = tf.Variable(
                    tf.random.uniform(
                        shape=[],
                        minval=range_[0],
                        maxval=range_[1],
                        dtype=self.dtype,
                    ),
                    trainable=trainable,
                )
        else:  # constant value
            # if name in self.bnd_dic:
            # value = self.bnd_dic[name].get_y2x(value)
            self.variables[name] = tf.Variable(
                value, dtype=self.dtype, trainable=trainable
            )
            self.init_val[name] = value

        if trainable:
            self.trainable_vars.append(name)

    def add_complex_var(
        self, name, polar=None, trainable=True, fix_vals=(1.0, 0.0)
    ):
        """
        Add a complex variable. Two real variables named **name+'r'** and **name+'i'** will be added into
        **self.variables**. The initial values will be given automatically according to its form of coordinate.

        :param name: The name of the complex variable.
        :param polar: Boolean. If it's **True**, **name+'r'** and **name+'i'** are defined in polar coordinate; otherwise they are defined in Cartesian coordinate.
        :param trainable: Boolean. If it's **True**, real variables **name+'r'** and **name+'i'** will be trainable.
        :param fix_vals: Length-2 array. If **trainable=False**, the fixed values for **name+'r'** and **name+'i'** are **fix_vals[0]**, **fix_vals[1]** respectively.
        """
        if polar is None:
            polar = self.polar
        var_r = name + "r"
        var_i = name + "i"
        if trainable:
            if polar:
                self.add_real_var(name=var_r, range_=(0, 2.0))
                self.add_real_var(name=var_i, range_=(-np.pi, np.pi))
            else:
                self.add_real_var(name=var_r, range_=(-1, 1))
                self.add_real_var(name=var_i, range_=(-1, 1))
        else:
            self.add_real_var(name=var_r, value=fix_vals[0], trainable=False)
            self.add_real_var(name=var_i, value=fix_vals[1], trainable=False)
        self.complex_vars[name] = polar

    def add_cartesiancp_var(
        self, name, polar=None, trainable=True, fix_vals=(1.0, 0.0, 0.0, 0.0)
    ):
        """
        Add a complex variable. Two real variables named **name+'r'** and **name+'i'** will be added into
        **self.variables**. The initial values will be given automatically according to its form of coordinate.

        :param name: The name of the complex variable.
        :param polar: Boolean. If it's **True**, **name+'r'** and **name+'i'** are defined in polar coordinate; otherwise they are defined in Cartesian coordinate.
        :param trainable: Boolean. If it's **True**, real variables **name+'r'** and **name+'i'** will be trainable.
        :param fix_vals: Length-4 array. If **trainable=False**, the fixed values for **name+'r'** and **name+'i'** are **fix_vals[0]**, **fix_vals[1]** respectively.
        """
        if polar is None:
            polar = self.polar
        # if chargeconjugate is None:
        # chargeconjugate = self.chargeconjugate
        var_r = name + "r"
        var_i = name + "i"
        var_deltar = name + "deltar"
        var_deltai = name + "deltai"
        if trainable:
            self.add_real_var(name=var_r, range_=(-1, 1))
            self.add_real_var(name=var_i, range_=(-1, 1))
            self.add_real_var(name=var_deltar, range_=(-1, 1))
            self.add_real_var(name=var_deltai, range_=(-1, 1))
        else:
            self.add_real_var(name=var_r, value=fix_vals[0], trainable=False)
            self.add_real_var(name=var_i, value=fix_vals[1], trainable=False)
            self.add_real_var(
                name=var_deltar, value=fix_vals[2], trainable=False
            )
            self.add_real_var(
                name=var_deltai, value=fix_vals[3], trainable=False
            )

        self.complex_vars[name] = polar

    def remove_var(self, name):
        """
        Remove a variable from **self.variables**. More specifically, two variables (**name+'r'** and **name+'i'**)
        will be removed if it's complex.

        :param name: The name of the variable
        """
        if name in self.complex_vars:
            del self.complex_vars[name]
            name_r = name + "r"
            name_i = name + "i"
            if self.variables[name_r].trainable:
                if name_r in self.trainable_vars:
                    self.trainable_vars.remove(name_r)
            if self.variables[name_i].trainable:
                if name_i in self.trainable_vars:
                    self.trainable_vars.remove(name_i)
            for l in self.same_list:
                if name_r in l:
                    l.remove(name_r)
                if name_i in l:
                    l.remove(name_i)
            if name_r in self.bnd_dic:
                del self.bnd_dic[name_r]
            if name_i in self.bnd_dic:
                del self.bnd_dic[name_i]
            del self.variables[name_r]
            del self.variables[name_i]
        else:
            if self.variables[name].trainable:
                if name in self.trainable_vars:
                    self.trainable_vars.remove(name)
            for l in self.same_list:
                if name in l:
                    l.remove(name)
            if name in self.bnd_dic:
                del self.bnd_dic[name]
            del self.variables[name]

    def rename_var(self, name, new_name, cplx=False):
        """
        Rename a variable.

        :param name: Name of the variable
        :param new_name: New name
        :param cplx: Boolean. Users should indicate if this variable is complex or not.
        """
        if cplx:
            self.complex_vars[new_name] = self.complex_vars[name]
            del self.complex_vars[name]
            name_r = name + "r"
            name_i = name + "i"
            new_name_r = new_name + "r"
            new_name_i = new_name + "i"
            if self.variables[name_r].trainable:
                if name_r in self.trainable_vars:
                    self.trainable_vars.remove(name_r)
                    self.trainable_vars.append(new_name_r)
            if self.variables[name_i].trainable:
                if name_i in self.trainable_vars:
                    self.trainable_vars.remove(name_i)
                    self.trainable_vars.append(new_name_i)
            for l in self.same_list:
                if name_r in l:
                    l.remove(name_r)
                    l.append(new_name_r)
                if name_i in l:
                    l.remove(name_i)
                    l.append(new_name_i)
            if name_r in self.bnd_dic:
                self.bnd_dic[new_name_r] = self.bnd_dic[name_r]
                del self.bnd_dic[name_r]
            if name_i in self.bnd_dic:
                self.bnd_dic[new_name_i] = self.bnd_dic[name_i]
                del self.bnd_dic[name_i]
            self.variables[new_name_r] = self.variables[name_r]
            del self.variables[name_r]
            self.variables[new_name_i] = self.variables[name_i]
            del self.variables[name_i]
        else:
            if self.variables[name].trainable:
                self.trainable_vars.remove(name)
                self.trainable_vars.append(new_name)
            for l in self.same_list:
                if name in l:
                    l.remove(name)
                    l.append(new_name)
            if name in self.bnd_dic:
                self.bnd_dic[new_name] = self.bnd_dic[name]
                del self.bnd_dic[name]
            self.variables[new_name] = self.variables[name]
            del self.variables[name]

    def refresh_vars(self, init_val=None, bound_dic=None):
        """
        Refresh all trainable variables
        """
        if bound_dic is None:
            bound_dic = self.bnd_dic
        if init_val is None:
            init_val = self.init_val
        cplx_vars = []
        for name in self.complex_vars:  # complex vars
            name_r = name + "r"
            name_i = name + "i"
            if self.complex_vars[name] == False:  # xy coordinate
                if name_r in self.trainable_vars:
                    cplx_vars.append(name_r)
                    self.variables[name_r].assign(
                        tf.random.uniform(
                            shape=[], minval=-1, maxval=1, dtype=self.dtype
                        )
                    )
                if name_i in self.trainable_vars:
                    cplx_vars.append(name_i)
                    self.variables[name_i].assign(
                        tf.random.uniform(
                            shape=[], minval=-1, maxval=1, dtype=self.dtype
                        )
                    )
            else:  # polar coordinate
                if name_r in self.trainable_vars:
                    cplx_vars.append(name_r)
                    self.variables[name_r].assign(
                        tf.random.uniform(
                            shape=[], minval=0, maxval=2, dtype=self.dtype
                        )
                    )
                if name_i in self.trainable_vars:
                    cplx_vars.append(name_i)
                    self.variables[name_i].assign(
                        tf.random.uniform(
                            shape=[],
                            minval=-np.pi,
                            maxval=np.pi,
                            dtype=self.dtype,
                        )
                    )
        # all_vars = set(self.trainable_vars) # real vars
        # real_vars = all_vars - set(cplx_vars)
        for name in set(init_val) & set(self.trainable_vars):
            if hasattr(init_val[name], "__len__"):
                mu = init_val[name][0]
                sigma = init_val[name][1]
                if name not in bound_dic:
                    val = tf.random.normal(
                        shape=[], mean=mu, stddev=sigma, dtype=self.dtype
                    )
                else:
                    range_ = bound_dic[name]
                    while True:
                        val = tf.random.normal(
                            shape=[], mean=mu, stddev=sigma, dtype=self.dtype
                        )
                        if val < range_[1] and val > range_[0]:
                            break
                self.variables[name].assign(val)
            else:
                if init_val[name] is not None:
                    self.variables[name].assign(init_val[name])

        for name in set(bound_dic) - set(init_val):
            _min, _max = bound_dic[name]
            if name not in self.trainable_vars:
                continue
            if _min is not None:
                if _max is not None:
                    val = tf.random.uniform(
                        shape=[], minval=_min, maxval=_max, dtype=self.dtype
                    )
                else:
                    val = _min + np.random.chisquare(df=1)
            else:
                if _max is not None:
                    val = _max - np.random.chisquare(df=1)
            self.variables[name].assign(val)

    def set_fix(self, name, value=None, unfix=False):
        """
        Fix or unfix a variable (change the trainability)
        :param name: The name of the variable
        :param value: The fixed value. It's useless if **unfix=True**.
        :param unfix: Boolean. If it's **True**, the variable will become trainable rather than be fixed.
        """
        if value is None:
            value = self.variables[name].value
        else:
            if name in self.bnd_dic:
                value = self.bnd_dic[name].get_y2x(value)
        var = tf.Variable(value, dtype=self.dtype, trainable=unfix)
        self.variables[name] = var
        if unfix:
            if name in self.trainable_vars:
                warnings.warn("{} has been freed already!".format(name))
            else:
                self.trainable_vars.append(name)
        else:
            if name in self.trainable_vars:
                self.trainable_vars.remove(name)
            else:
                warnings.warn("{} has been fixed already!".format(name))

    def set_bound(self, bound_dic, func=None, overwrite=False):
        """
        Set boundary for the trainable variables. The variables will be constrained in their ranges while fitting.

        :param bound_dic: Dictionary. E.g. **{"name1":(-1.0,1.0), "name2":(None,1.0)}**. In this example, **None** means it has no lower limit.
        :param func: String. Users can provide a string to describe the transforming function. For details, refer to class **tf_pwa.variable.Bound**.
        :param overwrite: Boolean. If it's ``True``, the program will not throw a warning when overwrite a variable with the same name.
        """
        for name in bound_dic:
            if name in self.bnd_dic:
                if not overwrite:
                    warnings.warn("Overwrite bound of {}!".format(name))
            self.bnd_dic[name] = Bound(*bound_dic[name], func=func)
            if name in self.variables:
                has_same = False
                for i in self.same_list:
                    if name in i[1:]:
                        has_same = True
                        break
                if has_same:
                    continue
                # val = self.get(name).numpy()
                # self.set(name, self.bnd_dic[name].get_y2x(val))

    def _remove_bound(self, name):
        if name in self.variables:
            has_same = False
            for i in self.same_list:
                if name in i[1:]:
                    has_same = True
                    break
            if not has_same:
                value = self.get(name, val_in_fit=False)
                # self.set(name, value)
        del self.bnd_dic[name]

    def remove_bound(self):
        """
        Remove a boundary for a variable
        """
        bnd_dic = self.bnd_dic.copy()
        for i in bnd_dic:
            self._remove_bound(i)
        return bnd_dic

    def set_share_r(self, name_list):  # name_list==[name1,name2,...]
        """
        If some complex variables want to share their radia variable while their phase variable are still different.
        Users can set this type of constrain using this method.

        :param name_list: List of strings. Note the strings should be the name of the complex variables rather than of their radium parts.
        """
        self.xy2rp_all(name_list)
        name_r_list = [name + "r" for name in name_list]
        self.set_same(name_r_list)
        for name in name_r_list:
            self.complex_vars[name[:-1]] = name_r_list

    def set_same(self, name_list, cplx=False):
        """
        Set some variables to be the same.

        :param name_list: List of strings. Name of the variables.
        :param cplx: Boolean. Whether the variables are complex or real.
        """
        tmp_list = []
        for name in name_list:
            for add_list in self.same_list:
                if name not in self.variables:
                    continue
                if name in add_list:
                    tmp_list += add_list
                    self.same_list.remove(add_list)
                    break

        for i in tmp_list:
            if i not in name_list:
                name_list.append(i)  # 去掉重复元素

        def same_real(name_list):
            name_list = [i for i in name_list if i in self.variables]
            if len(name_list) == 0:
                return
            var = self.variables[name_list[0]]
            for name in name_list[1:]:
                if name in self.trainable_vars:
                    self.trainable_vars.remove(name)
                else:
                    # if one is untrainable, the others will all be untrainable
                    var = self.variables.get(name, None)
                    if var is not None:
                        if name_list[0] in self.trainable_vars:
                            self.trainable_vars.remove(name_list[0])
            for name in name_list:
                self.variables[name] = var

        if cplx:
            same_real([name + "r" for name in name_list])
            same_real([name + "i" for name in name_list])
        else:
            same_real(name_list)
        self.same_list.append(name_list)

    def get(self, name, val_in_fit=True):
        """
        Get a real variable. If ``val_in_fit is True``, this is the variable used in fitting, not considering its boundary transformation.

        :param name: String
        :return: tf.Variable
        """
        if name not in self.variables:
            raise Exception("{} not found".format(name))
        if not val_in_fit or name not in self.bnd_dic:
            return self.variables[name].numpy()  # tf.Variable
        else:
            return self.bnd_dic[name].get_y2x(self.variables[name].numpy())

    def set(self, name, value, val_in_fit=True):
        """
        Set value for a real variable. If ``val_in_fit is True``, this is the variable used in fitting, not considering its boundary transformation.

        :param name: String
        :param value: Real number
        """
        if val_in_fit and name in self.bnd_dic:
            value = self.bnd_dic[name].get_x2y(value)
        if name in self.variables:
            self.variables[name].assign(value)
        else:
            warnings.warn("{} not found".format(name))

    def rp2xy(self, name):
        """
        Transform a complex variable into Cartesian coordinate.
        :param name: String
        """
        if self.complex_vars[name] != True:  # if not polar (already xy)
            return
        r = self.variables[name + "r"]
        p = self.variables[name + "i"]
        x = r * tf.cos(p)
        y = r * tf.sin(p)
        self.variables[name + "r"].assign(x)
        self.variables[name + "i"].assign(y)
        self.complex_vars[name] = False
        for l in self.same_list:
            if name in l:
                for i in l:
                    self.complex_vars[i] = False
                break

    def xy2rp(self, name):
        """
        Transform a complex variable into polar coordinate.
        :param name: String
        """
        if self.complex_vars[name] != False:  # if already polar
            return
        x = self.variables[name + "r"]
        y = self.variables[name + "i"]
        r = tf.sqrt(x * x + y * y)
        p = tf.atan2(y, x)
        self.variables[name + "r"].assign(r)
        self.variables[name + "i"].assign(p)
        self.complex_vars[name] = True
        for l in self.same_list:
            if name in l:
                for i in l:
                    self.complex_vars[i] = True
                break

    @property
    def trainable_variables(self):
        """
        List of tf.Variable. It is similar to **tf.keras.Model.trainable_variables**.
        """
        vars_list = []
        for name in self.trainable_vars:
            vars_list.append(self.variables[name])
        return vars_list

    def get_all_val(self, val_in_fit=False):  # if bound transf var
        """
        Get the values of all trainable variables.

        :param val_in_fit: Boolean. If it's **True**, the values will be the ones that are actually used in fitting (thus may not be the physical values because of the boundary transformation).
        :return: List of real numbers.
        """
        vals = []
        for name in self.trainable_vars:
            xval = self.get(name, val_in_fit)
            vals.append(xval)
        return vals  # list (for list of tf.Variable use self.trainable_variables; for dict of all vars, use self.variables)

    def get_all_dic(self, trainable_only=False):
        """
        Get a dictionary of all variables.

        :param trainable_only: Boolean. If it's **True**, the dictionary only contains the trainable variables.
        :return: Dictionary
        """
        # self.std_polar_all()
        dic = {}
        if trainable_only:
            for i in self.trainable_vars:
                val = self.variables[i].numpy()
                # if i in self.bnd_dic:
                #     val = self.bnd_dic[i].get_y2x(val)
                dic[i] = val
        else:
            for i in self.variables:
                val = self.variables[i].numpy()
                # if i in self.bnd_dic:
                #    val = self.bnd_dic[i].get_y2x(val)
                dic[i] = val
        return dic

    def set_all(self, vals, val_in_fit=False):  # use either dict or list
        """
        Set values for some variables.

        :param vals: It can be either a dictionary or a list of real numbers. If it's a list, the values correspond to all trainable variables in order.
        """
        if type(vals) == dict:
            for name in vals:
                self.set(name, vals[name], val_in_fit=val_in_fit)
        else:
            i = 0
            for name in self.trainable_vars:
                self.set(name, vals[i], val_in_fit)
                i += 1

    def rp2xy_all(self, name_list=None):
        """
        If **name_list** is not provided, this method will transform all complex variables into Cartesian coordinate.

        :param name_list: List of names of complex variables
        """
        if not name_list:
            name_list = self.complex_vars
        for name in name_list:
            self.rp2xy(name)
        self.polar = False

    def xy2rp_all(self, name_list=None):
        """
        If **name_list** is not provided, this method will transform all complex variables into polar coordinate.

        :param name_list: List of names of complex variables
        """
        if not name_list:
            name_list = self.complex_vars
        for name in name_list:
            self.xy2rp(name)
        self.polar = True

    @staticmethod
    def _std_polar_angle(p, a=-np.pi, b=np.pi):
        return (p - a) % (b - a) + a

    def std_polar(self, name):
        """
        Transform a complex variable into standard polar coordinate, which mean its radium part is positive, and its
        phase part is between :math:`-\\pi` to :math:`\\pi`.
        :param name: String
        """
        self.xy2rp(name)
        r = self.variables[name + "r"]
        p = self.variables[name + "i"]
        if r < 0:
            r.assign(tf.abs(r))
            if type(self.complex_vars[name]) == list:
                for name_r in self.complex_vars[name]:
                    self.variables[name_r[:-1] + "i"].assign_add(np.pi)
            else:
                p.assign_add(np.pi)
        self._std_polar_angle(p)

    def std_polar_all(self):  # std polar expression: r>0, -pi<p<pi
        """
        Transform all complex variables into standard polar coordinate.
        """
        for name in self.complex_vars:
            self.std_polar(name)

    def trans_params(self, polar):
        """
        Transform all complex variables into either polar coordinate or Cartesian coordinate.

        :param polar: Boolean
        """
        if polar:
            self.std_polar_all()
        else:
            self.rp2xy_all()

    def set_trans_var(self, xvals):
        """
        :math:`y = y(x)`

        :param fcn_grad: The return of class **tf_pwa.model**???
        :return:
        """

        xvals = np.array(xvals)
        yvals = xvals.copy()
        for i, name in enumerate(self.trainable_vars):
            if name in self.bnd_dic:
                yvals[i] = self.bnd_dic[name].get_x2y(xvals[i])
        self.set_all(yvals)

    def trans_fcn_grad(self, fcn_grad):  # bound transform fcn and grad
        """
        :math:`F(x)=F(y(x))`, :math:`G(x)=\\frac{dF}{dx}=\\frac{dF}{dy}\\frac{dy}{dx}`

        :param fcn_grad: The return of class **tf_pwa.model**???
        :return:
        """

        def fcn_t(xvals):
            xvals = np.array(xvals)
            yvals = xvals.copy()
            dydxs = []
            i = 0
            for name in self.trainable_vars:
                if name in self.bnd_dic:
                    yvals[i] = self.bnd_dic[name].get_x2y(xvals[i])
                    dydxs.append(self.bnd_dic[name].get_dydx(xvals[i]))
                else:
                    dydxs.append(1)
                i += 1
            fcn, grad_yv = fcn_grad(yvals)
            grad = np.array(grad_yv) * dydxs
            return fcn, grad

        return fcn_t

    def trans_grad_hessp(self, f):  # bound transform fcn and grad
        """
        :math:`F(x)=F(y(x))`, :math:`G(x)=\\frac{dF}{dx}=\\frac{dF}{dy}\\frac{dy}{dx}`

        :param fcn_grad: The return of class **tf_pwa.model**???
        :return:
        """

        def f_wrap(xvals, p):
            xvals = np.array(xvals)
            yvals = xvals.copy()
            dydxs = []
            dydxs2 = []
            i = 0
            for name in self.trainable_vars:
                if name in self.bnd_dic:
                    yvals[i] = self.bnd_dic[name].get_x2y(xvals[i])
                    dydxs.append(self.bnd_dic[name].get_dydx(xvals[i]))
                    dydxs2.append(self.bnd_dic[name].get_d2ydx2(xvals[i]))
                else:
                    dydxs.append(1)
                    dydxs2.append(0)
                i += 1

            dydxs = np.array(dydxs)
            dydxs2 = np.array(dydxs2)

            # print(xvals.shape, p.shape, dydxs.shape, len(self.trainable_vars))

            grad_yv, hessp_yv = f(yvals, p * dydxs)
            grad = np.array(grad_yv) * dydxs
            # print("trans", p, grad, hessp_yv * dydxs + grad_yv * dydxs2)
            return grad, hessp_yv * dydxs + grad_yv * dydxs2 * p

        return f_wrap

    def trans_f_grad_hess(self, f):  # bound transform fcn and grad
        """
        :math:`F(x)=F(y(x))`, :math:`G(x)=\\frac{dF}{dx}=\\frac{dF}{dy}\\frac{dy}{dx}`

        :param fcn_grad: The return of class **tf_pwa.model**???
        :return:
        """

        def f_wrap(xvals):
            xvals = np.array(xvals)
            yvals = xvals.copy()
            dydxs = []
            dydxs2 = []
            i = 0
            for name in self.trainable_vars:
                if name in self.bnd_dic:
                    yvals[i] = self.bnd_dic[name].get_x2y(xvals[i])
                    dydxs.append(self.bnd_dic[name].get_dydx(xvals[i]))
                    dydxs2.append(self.bnd_dic[name].get_d2ydx2(xvals[i]))
                else:
                    dydxs.append(1)
                    dydxs2.append(0)
                i += 1

            dydxs = np.array(dydxs)
            dydxs2 = np.array(dydxs2)

            # print(yvals) # , xvals.shape, p.shape, dydxs.shape, len(self.trainable_vars))

            f2, grad_yv, hessp_yv = f(yvals)
            grad = np.array(grad_yv) * dydxs
            # print("trans", p, grad, hessp_yv * dydxs + grad_yv * dydxs2)
            return (
                f2,
                grad,
                dydxs[:, None] * hessp_yv * dydxs[None, :]
                + np.diag(grad_yv * dydxs2),
            )

        return f_wrap

    def trans_error_matrix(self, hess_inv, xvals):
        """
        Bound trans for error matrix
        :math:`F(x)=F(y(x))`, :math:`V_y = y' V_x y'`

        :return:
        """
        xvals = np.array(xvals)
        dydxs = []
        for i, name in enumerate(self.trainable_vars):
            if name in self.bnd_dic:
                dydxs.append(self.bnd_dic[name].get_dydx(xvals[i]))
            else:
                dydxs.append(1)
        dydx = np.array(dydxs)
        hess_inv = dydx[:, None] * np.array(hess_inv) * dydx[None, :]
        return hess_inv

    @contextlib.contextmanager
    def error_trans(self, err_matrix):
        with ParamsTrans(self, err_matrix).trans() as f:
            yield f

    def minimize(self, fcn, jac=True, method="BFGS", mini_kwargs={}):
        """
        minimize a give function
        """
        if hasattr(fcn, "nll_grad"):
            f = fcn.nll_grad
        else:

            def f(x):
                self.set_all(x)
                with tf.GradientTape() as tape:
                    y = fcn()
                g = tape.gradient(
                    y, self.trainable_variables, unconnected_gradients="zero"
                )
                return float(y), np.array([float(i) for i in g])

        x0 = self.get_all_val(True)

        f2 = self.trans_fcn_grad(f)
        if isinstance(method, str):
            from scipy.optimize import minimize as mini

            if jac != True:

                def f(x):
                    self.set_all(x)
                    y = fcn()
                    return float(y)

                ret = mini(
                    f, np.array(x0), jac=jac, method=method, **mini_kwargs
                )
            else:
                ret = mini(
                    f2, np.array(x0), jac=jac, method=method, **mini_kwargs
                )
        else:
            ret = method(f2, np.array(x0), **mini_kwargs)
        self.set_all(ret.x, val_in_fit=True)
        ret.x = np.array(self.get_all_val())
        ret.hess_inv = self.trans_error_matrix(ret.hess_inv, ret.x)
        return ret


class Bound(object):
    """
    This class provides methods to implement the boundary constraint for a variable.
    It has dependence on `SymPy <https://www.sympy.org/en/index.html>`_ .
    The boundary-transforming function can transform a variable *x* defined in the real domain to a variable *y* defined
    in a limited range *(a,b)*. *y* should be the physical parameter but *x* is the one used while fitting.

    :param a: Real number. The lower boundary
    :param b: Real number. The upper boundary
    :param func: String. The boundary-transforming function. By default, if neither **a** or **b** is **None**, **func** is **"(b-a)*(sin(x)+1)/2+a"**; else if only **a** is **None**, **func** is **"b+1-sqrt(x**2+1)"**; else if only **b** is **None**, **func** is **"a-1+sqrt(x**2+1)"**; else **func** is **"x"**.

    **a**, **b**, **func** can be refered by **self.lower**, **self.upper**, **self.func**.
    """

    def __init__(self, a=None, b=None, func=None):
        if a is not None and b is not None and a > b:
            raise Exception("Lower bound is larger than upper bound!")
        self.lower = a
        self.upper = b
        if func:
            self.func = func  # from R (x) to a limited range (y) #Note: y is gls but x is the var in fitting
        else:
            if a is None:
                if b is None:
                    self.func = "x"
                else:
                    self.func = "b+1-sqrt(x**2+1)"
            else:
                if b is None:
                    self.func = "a-1+sqrt(x**2+1)"
                else:
                    self.func = "(b-a)*(sin(x)+1)/2+a"
        self.f, self.df, self.df2, self.inv = self.get_func()

    def __repr__(self):
        return "[" + str(self.lower) + ", " + str(self.upper) + "]"

    def get_func(self):  # init func string into sympy f(x) or f(y)
        """
        Initialize the function string into **sympy** objects.

        :return: **sympy** objects **f**, **df**, **inv**, which are the function, its derivative and its inverse function.
        """
        x, a, b, y = sy.symbols("x a b y")
        f = sy.sympify(self.func)
        f = f.subs({a: self.lower, b: self.upper})
        df = sy.diff(f, x)
        df2 = sy.diff(df, x)
        inv = sy.solve(f - y, x)
        if hasattr(inv, "__len__"):
            inv = inv[-1]
        return f, df, df2, inv

    def get_x2y(self, val):  # var->gls
        """
        To derive *y* from *x*

        :param val: Real number *x*
        :return: Real number *y*
        """
        x = sy.symbols("x")
        return float(self.f.evalf(subs={x: val}))

    def get_y2x(self, val):  # gls->var
        """
        To derive *x* from *y*. *y* will be set to *a* if *y<a*, and *y* will be set to *b* if *y>b*.

        :param val: Real number *y*
        :return: Real number *x*
        """
        y = sy.symbols("y")
        if self.lower is not None and val < self.lower:
            val = self.lower
        elif self.upper is not None and val > self.upper:
            val = self.upper
        x = self.inv.evalf(subs={y: val})
        return complex(x).real

    def get_dydx(self, val):  # gradient in fitting: dNLL/dx = dNLL/dy * dy/dx
        """
        To calculate the derivative :math:`\\frac{dy}{dx}`.

        :param val: Real number *x*
        :return: Real number :math:`\\frac{dy}{dx}`
        """
        x = sy.symbols("x")
        return float(self.df.evalf(subs={x: val}))

    def get_d2ydx2(
        self, val
    ):  # gradient in fitting: dNLL/dx = dNLL/dy * dy/dx
        """
        To calculate the derivative :math:`\\frac{dy}{dx}`.

        :param val: Real number *x*
        :return: Real number :math:`\\frac{dy}{dx}`
        """
        x = sy.symbols("x")
        return float(self.df2.evalf(subs={x: val}))


def _get_val_from_index(val, index):
    for i in index:
        val = val[i]
    return val


def _shape_func(f, shape, name="", idx=[], **kwargs):
    if not shape:
        f(name, idx, **kwargs)
    else:
        for i in range(shape[0]):
            _shape_func(f, shape[1:], name + "_" + str(i), idx + [i], **kwargs)


regist_config("vm", VarsManager(dtype="float64"))


# vm = VarsManager(dtype=tf.float64)
class Variable(object):
    """
    This class has interface to **VarsManager**. It is convenient for users to define a group of real variables,
    since it may be more perceptually intuitive to define them together.

    By calling the instance of this class, it returns the value of this variable. The type is tf.Tensor.

    :param name: The name of the variable group
    :param shape: The shape of the group. E.g. for a 4*3*2 matrix, **shape** is **[4,3,2]**. By default, **shape** is [] for a real variable.
    :param cplx: Boolean. Whether the variable (or the variables) are complex or not.
    :param vm: VarsManager. It is by default the one automatically defined in the global scope by the program.
    :param overwrite: Boolean. If it's ``True``, the program will not throw a warning when overwrite a variable with the same name.
    :param kwargs: Other arguments that may be used when calling **self.real_var()** or **self.cplx_var()**
    """

    def __init__(
        self,
        name,
        shape=None,
        cplx=False,
        vm=None,
        overwrite=True,
        is_cp=False,
        **kwargs
    ):
        if shape is None:
            shape = []
        if vm is None:
            vm = get_config("vm")
        self.vm = vm
        self.name = name
        for i in self.vm.var_head:
            if i.name == self.name:
                if not overwrite:
                    ex = 1
                    while True:
                        ex_name = i.name + "_" + str(ex)
                        judg = True
                        for ii in self.vm.var_head:
                            if ii.name == ex_name:
                                ex += 1
                                judg = False
                                break
                        if judg:
                            break
                    self.name = ex_name
                else:
                    for j in self.vm.var_head[i]:
                        self.vm.remove_var(j)
                    del self.vm.var_head[i]
                break
        self.vm.var_head[self] = []
        if type(shape) == int:
            shape = [shape]
        self.shape = shape
        self.cplx = cplx
        self.cp_effect = is_cp
        if is_cp:
            # print("Variable init cplx_cpvar" )
            self.cplx_cpvar(**kwargs)
        elif cplx:
            self.cplx_var(**kwargs)
        else:
            self.real_var(**kwargs)
        self.bound = None

    def real_var(self, value=None, range_=None, fix=False):
        """
        It implements interface to ``VarsManager.add_real_var()``, but supports variables that are not of non-shape.

        :param value: Real number. The value of all real components.
        :param range_: Length-2 array. The length of all real components.
        :param fix: Boolean. Whether the variable is fixed.
        """
        trainable = not fix

        def func(name, idx, **kwargs):
            self.vm.add_real_var(name, value, range_, trainable)
            self.vm.var_head[self].append(name)

        _shape_func(
            func,
            self.shape,
            self.name,
            value=value,
            range_=range_,
            trainable=trainable,
        )

    def cplx_var(self, polar=None, fix=False, fix_vals=(1.0, 0.0)):
        """
        It implements interface to ``VarsManager.add_complex_var()``, but supports variables that are not of non-shape.

        :param polar: Boolean. Whether the variable is defined in polar coordinate or in Cartesian coordinate.
        :param fix: Boolean. Whether the variable is fixed. It's enabled only if ``self.shape is None``.
        :param fix_vals: Length-2 tuple. The value of the fixed complex variable is ``fix_vals[0]+fix_vals[1]j``.
        """
        if not hasattr(fix_vals, "__len__"):
            fix_vals = [fix_vals, 0.0]

        def func(name, idx, **kwargs):
            trainable = not fix
            self.vm.add_complex_var(name, polar, trainable, fix_vals)
            self.vm.var_head[self].append(name)

        _shape_func(
            func, self.shape, self.name, polar=polar, fix_vals=fix_vals
        )

    def cplx_cpvar(
        self, polar=True, fix=False, fix_vals=(1.0, 0.0, 0.0, 0.0), value=0.0
    ):
        """
        It implements interface to ``VarsManager.add_complex_var()``, but supports variables that are not of non-shape.

        :param polar: Boolean. Whether the variable is defined in polar coordinate or in Cartesian coordinate.
        :param fix: Boolean. Whether the variable is fixed. It's enabled only if ``self.shape is None``.
        :param fix_vals: Length-4 tuple. The value of the fixed complex variable is ``fix_vals[0]+fix_vals[1]j``.
        """
        if not hasattr(fix_vals, "__len__"):
            fix_vals = [fix_vals, 0.0, 0.0, 0.0]

        def func(name, idx, **kwargs):
            trainable = not fix
            # print("Variable cplx_cpvar")
            self.vm.add_cartesiancp_var(name, polar, trainable, fix_vals)
            self.vm.var_head[self].append(name)

        _shape_func(
            func,
            self.shape,
            self.name,
            polar=polar,
            fix_vals=fix_vals,
            value=value,
        )

    def __repr__(self):
        return self.name

    @property
    def value(self):
        """
        :return: Ndarray of ``self.shape``.
        """
        return tf.Variable(self()).numpy()

    def set_value(self, value, index=None):
        if index is not None:
            assert len(index) == len(self.shape)
            var_name = self.name
            for i in index:
                var_name += "_" + str(index[i])
            if self.cp_effect == True:
                self.vm.set(var_name + "r", value[0])
                self.vm.set(var_name + "i", value[1])
                self.vm.set(var_name + "deltar", value[2])
                self.vm.set(var_name + "deltai", value[3])
            elif self.cplx == True:
                self.vm.set(var_name + "r", value[0])
                self.vm.set(var_name + "i", value[1])
            else:
                self.vm.set(var_name, value)
        else:
            value = np.array(value)

            def _get_val_from_index(val, index):
                for i in index:
                    val = val[i]
                return val

            if self.cp_effect == True:
                if value.shape[:-1] == ():

                    def func(name, idx, **kwargs):
                        self.vm.set(name + "r", value[0])
                        self.vm.set(name + "i", value[1])
                        self.vm.set(name + "deltar", value[2])
                        self.vm.set(name + "deltai", value[3])

                elif value.shape[:-1] == tuple(self.shape):

                    def func(name, idx, **kwargs):
                        self.vm.set(
                            name + "r", _get_val_from_index(value, idx)[0]
                        )
                        self.vm.set(
                            name + "i", _get_val_from_index(value, idx)[1]
                        )
                        self.vm.set(
                            name + "deltar", _get_val_from_index(value, idx)[2]
                        )
                        self.vm.set(
                            name + "deltai", _get_val_from_index(value, idx)[3]
                        )

                else:
                    raise Exception(
                        "The shape of value should be ", self.shape
                    )
            elif self.cplx == True:
                if value.shape[:-1] == ():

                    def func(name, idx, **kwargs):
                        self.vm.set(name + "r", value[0])
                        self.vm.set(name + "i", value[1])

                elif value.shape[:-1] == tuple(self.shape):

                    def func(name, idx, **kwargs):
                        self.vm.set(
                            name + "r", _get_val_from_index(value, idx)[0]
                        )
                        self.vm.set(
                            name + "i", _get_val_from_index(value, idx)[1]
                        )

                else:
                    raise Exception(
                        "The shape of value should be ", self.shape
                    )
            else:
                if value.shape == ():

                    def func(name, idx, **kwargs):
                        self.vm.set(name, value)

                elif value.shape == tuple(self.shape):

                    def func(name, idx, **kwargs):
                        self.vm.set(name, _get_val_from_index(value, idx))

                else:
                    raise Exception(
                        "The shape of value should be ", self.shape
                    )

            _shape_func(func, self.shape, self.name, idx=[])

    def set_rho(self, rho, index=None):
        if self.cplx is not True:
            raise Exception("This method only supports complex Variable!")
        if index is not None:
            assert len(index) == len(self.shape)
            var_name = self.name
            for i in index:
                var_name += "_" + str(index[i])
            self.vm.set(var_name + "r", rho)
        else:
            rho = np.array(rho)
            if rho.shape == ():

                def func(name, idx, **kwargs):
                    self.vm.set(name + "r", rho)

            elif rho.shape == tuple(self.shape):

                def func(name, idx, **kwargs):
                    self.vm.set(name + "r", _get_val_from_index(rho, idx)[0])

            else:
                raise Exception("The shape of rho should be ", self.shape)
            _shape_func(func, self.shape, self.name, idx=[])

    def set_phi(self, phi, index=None):
        if self.cplx is not True:
            raise Exception("This method only supports complex Variable!")
        if index is not None:
            assert len(index) == len(self.shape)
            var_name = self.name
            for i in index:
                var_name += "_" + str(index[i])
            self.vm.set(var_name + "i", phi)
        else:
            phi = np.array(phi)
            if phi.shape == ():

                def func(name, idx, **kwargs):
                    self.vm.set(name + "i", phi)

            elif phi.shape == tuple(self.shape):

                def func(name, idx, **kwargs):
                    self.vm.set(name + "i", _get_val_from_index(phi, idx)[0])

            else:
                raise Exception("The shape of phi should be ", self.shape)
            _shape_func(func, self.shape, self.name, idx=[])

    @property
    def variables(self):
        """
        Names of the real variables contained in this Variable instance.

        :return: List of string.
        """
        return self.vm.var_head[self]

    def rename(self, new_name):
        """Rename this Variable."""

        def func(name, idx):
            vn = self.name + name
            new_vn = new_name + name
            self.vm.rename_var(vn, new_vn, cplx=self.cplx)
            self.vm.var_head[self].remove(vn)
            self.vm.var_head[self].append(new_vn)

        _shape_func(func, self.shape, "")
        self.name = new_name

    def fixed(self, value=None):
        """
        Fix this Variable. Note only non-shape real Variable supports this method.

        :param value: Real number. The fixed value
        """
        if not self.shape:
            if self.cp_effect:
                if value is None:
                    value = [None, None, None, None]
                else:
                    cplx_value = complex(value)
                    value = [cplx_value.real, 0.0, 0.0, 0.0]
                    self.vm.set_fix(self.name + "r", value[0])
                    self.vm.set_fix(self.name + "i", value[1])
                    self.vm.set_fix(self.name + "deltar", value[2])
                    self.vm.set_fix(self.name + "deltai", value[3])
            elif self.cplx:
                if value is None:
                    value = [None, None]
                else:
                    cplx_value = complex(value)
                    value = [cplx_value.real, cplx_value.imag]
                self.vm.set_fix(self.name + "r", value[0])
                self.vm.set_fix(self.name + "i", value[1])
            else:
                self.vm.set_fix(self.name, value)
        else:
            raise Exception("Only shape==() real var supports 'fixed' method.")

    def freed(self):
        """
        Set free this Variable. Note only non-shape Variable supports this method.
        """
        if not self.shape:
            if self.cp_effect:
                self.vm.set_fix(self.name + "r", unfix=True)
                self.vm.set_fix(self.name + "i", unfix=True)
                self.vm.set_fix(self.name + "deltar", unfix=True)
                self.vm.set_fix(self.name + "deltai", unfix=True)
            elif self.cplx:
                self.vm.set_fix(self.name + "r", unfix=True)
                self.vm.set_fix(self.name + "i", unfix=True)
            else:
                self.vm.set_fix(self.name, unfix=True)
        else:
            raise Exception("Only shape==() var supports 'freed' method.")

    def set_fix_idx(self, fix_idx=None, fix_vals=None, free_idx=None):
        """
        :param fix_idx: Interger or list of integers. Which complex component in the innermost layer of the variable is fixed. E.g. If ``self.shape==[2,3,4]`` and ``fix_idx==[1,2]``, then Variable()[i][j][1] and Variable()[i][j][2] will be the fixed value.
        :param fix_vals: Float or length-2 float list for complex variable. The fixed value.
        :param free_idx: Interger or list of integers. Which complex component in the innermost layer of the variable is set free. E.g. If ``self.shape==[2,3,4]`` and ``fix_idx==[0]``, then Variable()[i][j][0] will be set free.
        """
        if not self.shape:
            raise Exception(
                "Only shape!=() var supports 'set_fix_idx' method to fix or free variables."
            )
        if free_idx is None:
            free_idx = []
        else:
            free_idx = free_idx % self.shape[-1]
        if fix_idx is None:
            fix_idx = []
        else:
            fix_idx = fix_idx % self.shape[-1]

        if not hasattr(fix_idx, "__len__"):
            fix_idx = [fix_idx]
        fix_idx_str = ["_" + str(i) for i in fix_idx]
        if not hasattr(free_idx, "__len__"):
            free_idx = [free_idx]
        free_idx_str = ["_" + str(i) for i in free_idx]

        if self.cp_effect:
            print(
                "I am cp_effect for set_fix_idx  fix_idx_str",
                fix_idx_str,
                " fix_vals ",
                fix_vals,
            )
            # print("I am cp_effect for set_fix_idx  free_idx_str", free_idx_str)
            if fix_vals is None:
                print("fix_vals is None", fix_vals)
                fix_vals = [None, None, None, None]
            elif not hasattr(fix_vals, "__len__"):
                fix_vals = [fix_vals, 0.0, fix_vals, 0.0]
            if len(fix_vals) < 4:
                fix_vals = (*fix_vals, 0, 0)
            print("fix_vals ", fix_vals)

            def func(name, idx):
                for ss in fix_idx_str:
                    if name.endswith(ss):
                        # print("set_fix_idx set name+r ", name)
                        self.vm.set_fix(name + "r", value=fix_vals[0])
                        self.vm.set_fix(name + "i", value=fix_vals[1])
                        if len(fix_vals) > 2:
                            self.vm.set_fix(name + "deltar", value=fix_vals[2])
                            self.vm.set_fix(name + "deltai", value=fix_vals[3])
                for ss in free_idx_str:
                    if name.endswith(ss):
                        self.vm.set_fix(name + "r", unfix=True)
                        self.vm.set_fix(name + "i", unfix=True)
                        self.vm.set_fix(name + "deltar", unfix=True)
                        self.vm.set_fix(name + "deltai", unfix=True)

            _shape_func(func, self.shape, self.name)
        elif self.cplx:
            if fix_vals is None:
                fix_vals = [None, None]
            elif not hasattr(fix_vals, "__len__"):
                fix_vals = [fix_vals, 0.0]

            def func(name, idx):
                for ss in fix_idx_str:
                    if name.endswith(ss):
                        self.vm.set_fix(name + "r", value=fix_vals[0])
                        self.vm.set_fix(name + "i", value=fix_vals[1])
                for ss in free_idx_str:
                    if name.endswith(ss):
                        self.vm.set_fix(name + "r", unfix=True)
                        self.vm.set_fix(name + "i", unfix=True)

            _shape_func(func, self.shape, self.name)
        else:

            def func(name, idx):
                for ss in fix_idx_str:
                    if name.endswith(ss):
                        self.vm.set_fix(name, value=fix_vals)
                for ss in free_idx_str:
                    if name.endswith(ss):
                        self.vm.set_fix(name, unfix=True)

            _shape_func(func, self.shape, self.name)

    def set_bound(self, bound, func=None, overwrite=False):
        """
        Set boundary for this Variable. Note only non-shape real Variable supports this method.

        :param bound: Length-2 tuple.
        :param func: String. Refer to class **tf_pwa.variable.Bound**.
        :param overwrite: Boolean. If it's ``True``, the program will not throw a warning when overwrite a variable with the same name.
        """
        if not self.shape:
            self.vm.set_bound({self.name: bound}, func, overwrite=overwrite)
            self.bound = self.vm.bnd_dic[self.name]
        else:
            raise Exception(
                "Only shape==() real var supports 'set_bound' method."
            )

    def set_same_ratio(self):
        assert self.cplx, "variable should be complex"

        var = []

        def func(name, idx):
            var.append(name)

        _shape_func(func, self.shape, self.name)
        self.vm.set_share_r(var)

    def r_shareto(self, Var):
        """
        Share the radium component to another Variable of the same shape. Only complex Variable supports this method.

        :param Var: Variable.
        """
        if self.shape != Var.shape:
            raise Exception("Shapes are not the same.")
        if not (self.cplx and Var.cplx):
            raise Exception("Type is not complex var.")

        def func(name, idx, **kwargs):
            self.vm.set_share_r([self.name + name, Var.name + name])

        _shape_func(func, self.shape, "")

    def sameas(self, Var):
        """
        Set the Variable to be the same with another Variable of the same shape.

        :param Var: Variable.
        """
        if self.shape != Var.shape:
            raise Exception("Shapes are not the same.")
        if self.cplx != Var.cplx:
            raise Exception("Types (real or complex) are not the same.")

        def func(name, idx, **kwargs):
            self.vm.set_same(
                [self.name + name, Var.name + name], cplx=self.cplx
            )

        _shape_func(func, self.shape, "")

    def __call__(self, charge=1):
        var_list = np.ones(shape=self.shape).tolist()
        if self.shape:

            def func(name, idx, **kwargs):
                tmp = var_list
                idx_str = name.split("_")[-len(self.shape) :]
                for i in idx_str[:-1]:
                    tmp = tmp[int(i)]

                if self.cp_effect:
                    if (name in self.vm.complex_vars) and self.vm.complex_vars[
                        name
                    ]:
                        real = (
                            self.vm.variables[name + "r"]
                            + charge * self.vm.variables[name + "deltar"]
                        )
                        imag = (
                            self.vm.variables[name + "i"]
                            + charge * self.vm.variables[name + "deltai"]
                        )
                        tmp[int(idx_str[-1])] = tf.complex(real, imag)
                elif self.cplx:
                    if (name in self.vm.complex_vars) and self.vm.complex_vars[
                        name
                    ]:
                        real = self.vm.variables[name + "r"] * tf.cos(
                            self.vm.variables[name + "i"]
                        )
                        imag = self.vm.variables[name + "r"] * tf.sin(
                            self.vm.variables[name + "i"]
                        )
                        tmp[int(idx_str[-1])] = tf.complex(real, imag)
                        # print("&&&&&pg",name)
                    else:
                        # print("$$$$$xg",name)
                        tmp[int(idx_str[-1])] = tf.complex(
                            self.vm.variables[name + "r"],
                            self.vm.variables[name + "i"],
                        )
                    # print(tmp[int(idx_str[-1])])
                else:
                    tmp[int(idx_str[-1])] = self.vm.variables[name]

            _shape_func(func, self.shape, self.name)
        else:
            if self.cplx:
                name = self.name
                if (name in self.vm.complex_vars) and self.vm.complex_vars[
                    name
                ]:
                    real = self.vm.variables[name + "r"] * tf.cos(
                        self.vm.variables[name + "i"]
                    )
                    imag = self.vm.variables[name + "r"] * tf.sin(
                        self.vm.variables[name + "i"]
                    )
                    var_list = tf.complex(real, imag)
                    # print("&&&pt",name)
                else:
                    # print("$$$xt",name)
                    var_list = tf.complex(
                        self.vm.variables[self.name + "r"],
                        self.vm.variables[self.name + "i"],
                    )
                # print(var_list)
            else:
                var_list = self.vm.variables[self.name]

        # return tf.stack(var_list)
        return var_list



import os
import warnings

import tensorflow as tf

# default configurations
# os.environ["CUDA_VISIBLE_DEVICES"] = "0"
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "1"
# os.environ["KMP_DUPLICATE_LIB_OK"]="TRUE" # for Mac

# pylint: disable=no-member
try:
    tf_version = int(str(tf.__version__).split(".")[0])
except Exception:
    tf_version = 2
if tf_version < 2:
    tf.compat.v1.enable_eager_execution()
    import tensorflow.compat.v2 as tf  # pragma pylint: disable=import-error


def set_gpu_mem_growth():
    gpus = tf.config.experimental.list_physical_devices("GPU")
    if gpus:
        try:
            # Currently, memory growth needs to be the same across GPUs
            for gpu in gpus:
                tf.config.experimental.set_memory_growth(gpu, True)
            logical_gpus = tf.config.experimental.list_logical_devices("GPU")
            # print(len(gpus), "Physical GPUs,", len(logical_gpus), "Logical GPUs")
        except RuntimeError as e:
            # Memory growth must be set before GPUs have been initialized
            print(e)


if "TF_PWA_GPU_FULL_MEM" in os.environ:
    if os.environ["TF_PWA_GPU_FULL_MEM"] == "0":
        set_gpu_mem_growth()
else:
    set_gpu_mem_growth()


class Module(object):
    pass


tensorflow_wrapper = Module()


def regist_function(name, var=None, base_mod=tensorflow_wrapper):
    mod = base_mod
    names = name.split(".")
    for i in names[:-1]:
        if not hasattr(mod, i):
            setattr(mod, i, Module())
        mod = getattr(mod, i)

    def wrapper(f):
        if hasattr(mod, names[-1]):
            warnings.warn("{} already exists.".format(name))
        setattr(mod, names[-1], f)
        return f

    if var is None:
        return wrapper
    else:
        return wrapper(var)


# @regist_function("cross", base_mod=tf)
def numpy_cross(a, b):
    a = tf.convert_to_tensor(a)
    b = tf.convert_to_tensor(b)
    a_0 = tf.zeros_like(b)
    b_0 = tf.zeros_like(a)
    a = a + a_0
    b = b + b_0
    # shape = tf.broadcast_static_shape(a.shape, b.shape)
    # a = tf.broadcast_to(a, shape)
    # b = tf.broadcast_to(b, shape)
    ret = tf.linalg.cross(a, b)
    return ret


# regist_function("sum", tf.reduce_sum, base_mod=tf)
regist_function("arctan2", tf.math.atan2, base_mod=tf)


# from .jax_wrapper import tf



import numpy as np

from tf_pwa.amp import Decay, DecayChain, DecayGroup, Particle
from tf_pwa.config_loader.data import load_data_mode
from tf_pwa.tests.test_full import gen_toy


def test_load_data(gen_toy):
    p = [Particle(f"name:{i}") for i in range(5)]
    dec = DecayGroup(
        [DecayChain([Decay(p[0], [p[1], p[2]]), Decay(p[1], [p[3], p[4]])])]
    )
    data_file = {
        "data": "toy_data/data.dat",
        "phsp": "toy_data/phsp.dat",
        "bg": "toy_data/bg.dat",
        "bg_weight": 0.1,
        "weight_scale": True,
    }
    data = load_data_mode(data_file, dec, "multi")
    w1 = data.get_data("bg")[0]["weight"]

    data = load_data_mode(data_file, dec, "simple")
    w2 = data.get_data("bg")["weight"]
    assert np.allclose(w1, w2)

    data_file["weight_scale"] = False
    data = load_data_mode(data_file, dec, "multi")
    w1 = data.get_data("bg")[0]["weight"]

    data = load_data_mode(data_file, dec, "simple")
    w2 = data.get_data("bg")["weight"]
    assert np.allclose(w1, w2)

    data = load_data_mode({**data_file, "cp_trans": True}, dec, "simple")
    charge = np.random.random(w2.shape) > 0.5
    np.savetxt("toy_data/test_charge.dat", charge)
    p1 = data.load_data("toy_data/bg.dat", charge="toy_data/test_charge.dat")
    assert np.sum(p1["charge_conjugation"] - charge) == 0



import os.path
import sys

from tf_pwa.config_loader import ConfigLoader
from tf_pwa.vis import draw_decay_struct

this_dir = os.path.dirname(__file__)
sys.path.insert(0, this_dir + "/..")


def main():
    config = ConfigLoader("config.yml")
    for i, dec in enumerate(config.get_decay()):
        draw_decay_struct(
            dec, filename="figure/fig_{}".format(i), format="png"
        )


if __name__ == "__main__":
    main()



"""
Examples for error propagation
------------------------------

Hear we use the same config in `particle_amplitude.py`

"""

config_str = """
decay:
    A:
       - [R1, B]
       - [R2, C]
       - [R3, D]
    R1: [C, D]
    R2: [B, D]
    R3: [B, C]

particle:
    $top:
       A: { mass: 1.86, J: 0, P: -1}
    $finals:
       B: { mass: 0.494, J: 0, P: -1}
       C: { mass: 0.139, J: 0, P: -1}
       D: { mass: 0.139, J: 0, P: -1}
    R1: [ R1_a, R1_b ]
    R1_a: { mass: 0.7, width: 0.05, J: 1, P: -1}
    R1_b: { mass: 0.5, width: 0.05, J: 0, P: +1}
    R2: { mass: 0.824, width: 0.05, J: 0, P: +1}
    R3: { mass: 0.824, width: 0.05, J: 0, P: +1}

"""

import matplotlib.pyplot as plt
import yaml

from tf_pwa.config_loader import ConfigLoader
from tf_pwa.histogram import Hist1D

config = ConfigLoader(yaml.full_load(config_str))
input_params = {
    "A->R1_a.BR1_a->C.D_total_0r": 6.0,
    "A->R1_b.BR1_b->C.D_total_0r": 1.0,
    "A->R2.CR2->B.D_total_0r": 2.0,
    "A->R3.DR3->B.C_total_0r": 1.0,
}
config.set_params(input_params)

data = config.generate_toy(1000)
phsp = config.generate_phsp(10000)


# %%
# After we calculated the parameters error, we will have an error matrix `config.inv_he` (using the inverse hessain).
# It is possible to save such matrix directly by `numpy.save` and to load it by `numpy.load`.

config.get_params_error(data=[data], phsp=[phsp])

# %%
# We can use the following method to profamance the error propagation
#
# .. math::
#    \sigma_{f} = \sqrt{\frac{\partial f}{\partial x_i} V_{ij} \frac{\partial f}{\partial x_j }}
#
# by adding some calculation here. We need to use `tensorflow` functions instead of those of `math` or `numpy`.
#

import tensorflow as tf

with config.params_trans() as pt:
    a2_r = pt["A->R2.CR2->B.D_total_0r"]
    a2_i = pt["A->R2.CR2->B.D_total_0r"]
    a2_x = a2_r * tf.cos(a2_i)

# %%
# And then we can calculate the error we needed as

print(a2_x.numpy(), pt.get_error(a2_x).numpy())

# %%
# We can also calculate some more complex examples, such as the ratio in mass range (0.75, 0.85) over full phace space.
# Even further, we can get the error of error in the meaning of error propagation.

from tf_pwa.data import data_mask

m_R2 = phsp.get_mass("(B, D)")
cut_cond = (m_R2 < 0.85) & (m_R2 > 0.75)

amp = config.get_amplitude()

with config.params_trans() as pt1:
    with config.params_trans() as pt:
        int_mc = tf.reduce_sum(amp(phsp))
        cut_phsp = data_mask(phsp, cut_cond)
        cut_int_mc = tf.reduce_sum(amp(cut_phsp))
        ratio = cut_int_mc / int_mc
    error = pt.get_error(ratio)

print(ratio.numpy(), "+/-", error.numpy())
print(error.numpy(), "+/-", pt1.get_error(error).numpy())



import numpy as np


def dat2npz(in_file, out_file):
    data = np.loadtxt(in_file)
    np.savez(out_file, data)


def main():
    import argparse

    parser = argparse.ArgumentParser(description="")
    parser.add_argument("in_file")
    parser.add_argument("out_file")
    results = parser.parse_args()
    dat2npz(results.in_file, results.out_file)


if __name__ == "__main__":
    main()



import subprocess

from tf_pwa.main import main, regist_subcommand


@regist_subcommand()
def sss(i: int, *args, j="ss"):
    print(i, type(i))
    return i


def test_cmd():
    ret = subprocess.getoutput("python -m tf_pwa")


def test_main_f():
    ret = main(["sss", "2", "--j=dd"])
    assert ret == 2



#!/usr/bin/env python3
import copy
import json
import os.path
import sys
from pprint import pprint

import numpy as np
import yaml

from tf_pwa.config_loader import ConfigLoader
from tf_pwa.data import data_index, data_mask, data_to_numpy
from tf_pwa.phasespace import PhaseSpaceGenerator

this_dir = os.path.dirname(__file__)
sys.path.insert(0, this_dir + "/..")


def gen_phasespace(top, finals, number):
    a = PhaseSpaceGenerator(top, finals)
    flat_mc_data = a.generate(number)
    return flat_mc_data


def simple_select(phsp, amp):
    weights = amp(phsp)
    max_weights = np.max(weights) * 1.01

    rnd = np.random.random(weights.shape)
    select_index = weights / max_weights > rnd
    select_data = data_mask(phsp, select_index)
    return select_data


def save_dat(file_name, data, config):
    idx = [("particle", i, "p") for i in config.get_dat_order(True)]
    data = data_to_numpy(data)
    data_p = np.array([data_index(data, i) for i in idx])

    dat = np.transpose(data_p, (1, 0, 2)).reshape((-1, 4))
    np.savetxt(file_name, dat)
    dat = np.loadtxt(file_name)


def gen_toy(config, params, file_name):
    config = ConfigLoader(config)
    try:
        config.set_params(params)
    except Exception as e:
        print(e)

    print("using params")
    print(json.dumps(config.get_params(), indent=2))
    amp = config.get_amplitude()
    phsp = config.get_data("phsp")

    select_data = simple_select(phsp, amp)

    save_dat(file_name, select_data, config)


def main():
    import argparse

    parser = argparse.ArgumentParser(description="calculate significance")
    parser.add_argument("--config", default="config.yml", dest="config")
    parser.add_argument("--params", default="final_params.json", dest="params")
    parser.add_argument("--output", default="data/toy.dat", dest="output")
    results = parser.parse_args()
    gen_toy(results.config, results.params, results.output)


if __name__ == "__main__":
    main()



import contextlib
import copy
import functools
import itertools
import json
import os
import re
import time
import warnings

import matplotlib.pyplot as plt
import numpy as np
import sympy as sy
import yaml
from scipy.interpolate import UnivariateSpline, interp1d
from scipy.optimize import BFGS, basinhopping, minimize

from tf_pwa.adaptive_bins import AdaptiveBound, cal_chi2
from tf_pwa.amp import (
    AmplitudeModel,
    DecayChain,
    DecayGroup,
    HelicityDecay,
    get_decay,
    get_particle,
)
from tf_pwa.applications import (
    cal_hesse_correct,
    cal_hesse_error,
    corr_coef_matrix,
    fit,
    fit_fractions,
    force_pos_def,
    num_hess_inv_3point,
)
from tf_pwa.cal_angle import prepare_data_from_decay
from tf_pwa.data import (
    data_index,
    data_merge,
    data_shape,
    data_split,
    data_to_numpy,
    load_data,
    save_data,
)
from tf_pwa.fit import FitResult
from tf_pwa.fit_improve import minimize as my_minimize
from tf_pwa.model import FCN, CombineFCN, MixLogLikehoodFCN, Model, Model_new
from tf_pwa.model.cfit import Model_cfit, Model_cfit_cached
from tf_pwa.model.opt_int import ModelCachedAmp, ModelCachedInt
from tf_pwa.particle import split_particle_type
from tf_pwa.root_io import has_uproot, save_dict_to_root
from tf_pwa.tensorflow_wrapper import tf
from tf_pwa.utils import time_print
from tf_pwa.variable import Variable, VarsManager

from .base_config import BaseConfig
from .data import load_data_mode
from .decay_config import DecayConfig


class ConfigLoader(BaseConfig):
    """class for loading config.yml"""

    def __init__(self, file_name, vm=None, share_dict=None):
        if share_dict is None:
            share_dict = {}
        super().__init__(file_name, share_dict)
        self.config["data"] = self.config.get("data", {})
        self.share_dict = share_dict
        self.decay_config = DecayConfig(self.config, share_dict)
        self.dec = self.decay_config.dec
        self.particle_map, self.particle_property = (
            self.decay_config.particle_map,
            self.decay_config.particle_property,
        )
        self.top, self.finals = self.decay_config.top, self.decay_config.finals
        self.full_decay = self.decay_config.full_decay
        self.decay_struct = self.decay_config.decay_struct
        if vm is None:
            vm = VarsManager()
        self.vm = vm
        self.amps = {}
        self.cached_data = None
        self.bound_dic = {}
        self.gauss_constr_dic = {}
        self.init_value = {}
        self.plot_params = PlotParams(
            self.config.get("plot", {}), self.decay_struct
        )
        self._neglect_when_set_params = []
        self.data = load_data_mode(self["data"], self.decay_struct)
        self.inv_he = None
        self._Ngroup = 1
        self.cached_fcn = {}
        self.extra_constrains = {}
        self.resolution_size = self.config.get("data", {}).get(
            "resolution_size", 1
        )

    @staticmethod
    def load_config(file_name, share_dict={}):
        if isinstance(file_name, dict):
            return copy.deepcopy(file_name)
        if isinstance(file_name, str):
            if file_name in share_dict:
                return ConfigLoader.load_config(share_dict[file_name])
            with open(file_name) as f:
                ret = yaml.load(f, yaml.FullLoader)
            return ret
        raise TypeError("not support config {}".format(type(file_name)))

    def get_data_file(self, idx):
        if idx in self.config["data"]:
            ret = self.config["data"][idx]
        else:
            ret = None
        return ret

    def get_dat_order(self, standard=False):
        order = self.config["data"].get("dat_order", None)
        if order is None:
            order = list(self.decay_struct.outs)
        else:
            order = [get_particle(str(i)) for i in order]
        if not standard:
            return order

        re_map = self.decay_struct.get_chains_map()

        def particle_item():
            for j in re_map:
                for k, v in j.items():
                    for s, l in v.items():
                        yield s, l

        new_order = []
        for i in order:
            for s, l in particle_item():
                if str(l) == str(i):
                    new_order.append(s)
                    break
            else:
                new_order.append(i)
        return new_order

    def get_data_mode(self):
        data_config = self.config.get("data", {})
        data = data_config.get("data", "")
        if isinstance(data, str):
            mode = "single"
        elif isinstance(data, list):
            data_i = data[0]
            if isinstance(data_i, str):
                return "single"
            elif isinstance(data_i, list):
                return "multi"

    @functools.lru_cache()
    def get_data(self, idx):
        return self.data.get_data(idx)

    def load_cached_data(self, file_name=None):
        return self.data.load_cached_data(file_name)

    def save_cached_data(self, data, file_name=None):
        self.data.save_cached_data(data, file_name=file_name)

    def get_all_data(self):
        datafile = ["data", "phsp", "bg", "inmc"]
        self.load_cached_data()
        data, phsp, bg, inmc = [self.get_data(i) for i in datafile]
        self._Ngroup = len(data)
        assert len(phsp) == self._Ngroup
        if bg is None:
            bg = [None] * self._Ngroup
        if inmc is None:
            inmc = [None] * self._Ngroup
        assert len(bg) == self._Ngroup
        assert len(inmc) == self._Ngroup
        self.save_cached_data(dict(zip(datafile, [data, phsp, bg, inmc])))
        return data, phsp, bg, inmc

    def get_data_index(self, sub, name):
        return self.plot_params.get_data_index(sub, name)

    def get_phsp_noeff(self):
        if "phsp_noeff" in self.config["data"]:
            phsp_noeff = self.get_data("phsp_noeff")
            assert len(phsp_noeff) == 1
            return phsp_noeff[0]
        warnings.warn(
            "No data file as 'phsp_noeff', using the first 'phsp' file instead."
        )
        return self.get_data("phsp")[0]

    def get_phsp_plot(self):
        if "phsp_plot" in self.config["data"]:
            assert len(self.config["data"]["phsp_plot"]) == len(
                self.config["data"]["phsp"]
            )
            return self.get_data("phsp_plot")
        return self.get_data("phsp")

    def get_decay(self, full=True):
        if full:
            return self.full_decay
        else:
            return self.decay_struct

    @functools.lru_cache()
    def get_amplitude(self, vm=None, name=""):
        use_tf_function = self.config.get("data", {}).get(
            "use_tf_function", False
        )
        decay_group = self.full_decay
        self.check_valid_jp(decay_group)
        if vm is None:
            vm = self.vm
        if vm in self.amps:
            return self.amps[vm]
        amp = AmplitudeModel(
            decay_group, vm=vm, name=name, use_tf_function=use_tf_function
        )
        self.add_constraints(amp)
        self.amps[vm] = amp
        return amp

    def eval_amplitude(self, *p):
        data = self.data.cal_angle(p)
        amp = self.get_amplitude()
        return amp(data)

    def check_valid_jp(self, decay_group):
        for decay_chain in decay_group:
            for dec in decay_chain:
                if isinstance(dec, HelicityDecay):
                    dec.check_valid_jp()

    def add_constraints(self, amp):
        constrains = self.config.get("constrains", {})
        if constrains is None:
            constrains = {}
        self.add_decay_constraints(amp, constrains.get("decay", {}))
        self.add_particle_constraints(amp, constrains.get("particle", {}))
        self.add_fix_var_constraints(amp, constrains.get("fix_var", {}))
        self.add_free_var_constraints(amp, constrains.get("free_var", []))
        self.add_var_range_constraints(amp, constrains.get("var_range", {}))
        self.add_var_equal_constraints(amp, constrains.get("var_equal", []))
        for k, v in self.extra_constrains.items():
            v(amp, constrains.get(k, {}))

    def register_extra_constrains(self, name, f=None):
        """
        add extra_constrains
        """

        def _reg(g):
            self.extra_constrains[name] = g
            return g

        if f is None:
            return _reg
        else:
            return _reg(f)

    def add_fix_var_constraints(self, amp, dic=None):
        if dic is None:
            dic = {}
        for k, v in dic.items():
            print("fix var: ", k, "=", v)
            amp.vm.set_fix(k, v)

    def add_free_var_constraints(self, amp, dic=None):
        if dic is None:
            dic = []
        for k in dic:
            print("free var: ", k)
            amp.vm.set_fix(k, unfix=True)

    def add_var_range_constraints(self, amp, dic=None):
        if dic is None:
            dic = {}
        for k, v in dic.items():
            print("variable range: ", k, " in ", v)
            self.bound_dic[k] = v

    def add_var_equal_constraints(self, amp, dic=None):
        if dic is None:
            dic = []
        for k in dic:
            print("same value:", k)
            amp.vm.set_same(k)

    def add_decay_constraints(self, amp, dic=None):
        if dic is None:
            dic = {}
        fix_total_idx = dic.get("fix_chain_idx", 0)
        fix_total_val = dic.get("fix_chain_val", np.random.uniform(0, 2))

        fix_decay = amp.decay_group.get_decay_chain(fix_total_idx)
        # fix which total factor
        fix_decay.total.set_fix_idx(fix_idx=0, fix_vals=(fix_total_val, 0.0))

    def add_particle_constraints(self, amp, dic=None):
        if dic is None:
            dic = {}

        res_dec = {}
        for d in amp.decay_group:
            for p_i in d.inner:
                i = str(p_i)
                res_dec[i] = d
                prefix_map = {
                    "m0": "mass",
                    "g0": "width",
                    "m_": "mass_",
                    "g_": "width_",
                }
                particle_config = self.decay_config.particle_property[i]

                params_dic = particle_config.get("params", None)
                if params_dic is None:
                    params_dic = {}
                for name in list(particle_config):
                    for prefix_i in prefix_map.keys():
                        if name.startswith(prefix_i):
                            name2 = (
                                prefix_map[prefix_i] + name[len(prefix_i) :]
                            )
                            params_dic[name2] = particle_config[name]
                    for prefix_i in prefix_map.values():
                        if name.startswith(prefix_i):
                            params_dic[name] = particle_config[name]

                variable_prefix = p_i.get_variable_name()

                set_prefix_constrains(self.vm, p_i, params_dic, self)

                simple_map = {"m": "mass", "g": "width"}

                gauss_constr = particle_config.get("gauss_constr", None)
                if gauss_constr is not None:
                    assert isinstance(gauss_constr, dict)
                    for k, v in gauss_constr.items():
                        if v:
                            name = simple_map.get(k, k)
                            full_name = variable_prefix + name
                            var0 = self.vm.get(full_name)
                            self.gauss_constr_dic[full_name] = (
                                var0.value,
                                v,
                            )
                        else:
                            raise Exception(
                                f"Need sigma of {k} of {p_i} when adding gaussian constraint"
                            )

                if isinstance(p_i.mass, Variable) or isinstance(
                    p_i.width, Variable
                ):
                    if "float" in particle_config and particle_config["float"]:
                        if "m" in particle_config["float"]:
                            p_i.mass.freed()  # set_fix(i+'_mass',unfix=True)
                            if "m_max" in particle_config:
                                upper = particle_config["m_max"]
                            # elif m_sigma is not None:
                            #    upper = self.config["particle"][i]["m0"] + 10 * m_sigma
                            else:
                                upper = None
                            if "m_min" in particle_config:
                                lower = particle_config["m_min"]
                            # elif m_sigma is not None:
                            #    lower = self.config["particle"][i]["m0"] - 10 * m_sigma
                            else:
                                lower = None
                            self.bound_dic[str(p_i.mass)] = (lower, upper)
                        else:
                            self._neglect_when_set_params.append(str(p_i.mass))
                        if "g" in particle_config["float"]:
                            p_i.width.freed()  # amp.vm.set_fix(i+'_width',unfix=True)
                            if "g_max" in particle_config:
                                upper = particle_config["g_max"]
                            # elif g_sigma is not None:
                            #    upper = self.config["particle"][i]["g0"] + 10 * g_sigma
                            else:
                                upper = None
                            if "g_min" in particle_config:
                                lower = particle_config["g_min"]
                            # elif g_sigma is not None:
                            #    lower = self.config["particle"][i]["g0"] - 10 * g_sigma
                            else:
                                lower = None
                            self.bound_dic[str(p_i.width)] = (lower, upper)
                        else:
                            self._neglect_when_set_params.append(
                                str(p_i.width)
                            )
                    else:
                        self._neglect_when_set_params.append(
                            i + "_mass"
                        )  # p_i.mass.name
                        self._neglect_when_set_params.append(
                            i + "_width"
                        )  # p_i.width.name

                # share helicity variables
                if "coef_head" in particle_config:
                    coef_head = particle_config["coef_head"]
                    if coef_head in res_dec:
                        d_coef_head = res_dec[coef_head]
                        for j, h in zip(d, d_coef_head):
                            if i in [str(jj) for jj in j.outs] or i is str(
                                j.core
                            ):
                                h.g_ls.sameas(j.g_ls)
                        # share total radium
                        d_coef_head.total.r_shareto(d.total)
                    else:
                        particle_config["coef_head"] = i

        equal_params = dic.get("equal", {})
        for k, v in equal_params.items():
            for vi in v:
                a = []
                for i in amp.decay_group.resonances:
                    if str(i) in vi:
                        a.append(i)
                a0 = a.pop(0)
                arg = getattr(a0, k)
                for i in a:
                    arg_i = getattr(i, k)
                    if isinstance(arg_i, Variable):
                        arg_i.sameas(arg)

    @functools.lru_cache()
    def _get_model(self, vm=None, name=""):
        amp = self.get_amplitude(vm=vm, name=name)
        model_name = self.config["data"].get("model", "auto")
        w_bkg, w_inmc = self._get_bg_weight()
        model = []
        if model_name == "cfit":
            bg_function = self.config["data"].get("bg_function", None)
            eff_function = self.config["data"].get("eff_function", None)
            w_bkg = self.config["data"]["bg_frac"]
            if not isinstance(w_bkg, list):
                w_bkg = [w_bkg]
            for wb in w_bkg:
                if self.config["data"].get("cached_amp", False):
                    model.append(
                        Model_cfit_cached(amp, wb, bg_function, eff_function)
                    )
                else:
                    model.append(
                        Model_cfit(amp, wb, bg_function, eff_function)
                    )
        elif "inmc" in self.config["data"]:
            float_wmc = self.config["data"].get(
                "float_inmc_ratio_in_pdf", False
            )
            if not isinstance(float_wmc, list):
                float_wmc = [float_wmc] * self._Ngroup
            assert len(float_wmc) == self._Ngroup
            for wb, wi, fw in zip(w_bkg, w_inmc, float_wmc):
                model.append(Model_new(amp, wb, wi, fw))
        elif self.config["data"].get("cached_int", False):
            for wb in w_bkg:
                model.append(ModelCachedInt(amp, wb))
        elif self.config["data"].get("cached_amp", False):
            for wb in w_bkg:
                model.append(ModelCachedAmp(amp, wb))
        else:
            for wb in w_bkg:
                model.append(
                    Model(amp, wb, resolution_size=self.resolution_size)
                )
        return model

    def _get_bg_weight(self, data=None, bg=None, display=True):
        w_bkg = self.config["data"].get("bg_weight", 0.0)
        if not isinstance(w_bkg, list):
            w_bkg = [w_bkg] * self._Ngroup
        assert len(w_bkg) == self._Ngroup
        w_inmc = self.config["data"].get("inject_ratio", 0.0)
        if not isinstance(w_inmc, list):
            w_inmc = [w_inmc] * self._Ngroup
        assert len(w_inmc) == self._Ngroup
        weight_scale = self.config["data"].get("weight_scale", False)  # ???
        if weight_scale:
            data = data if data is not None else self.get_data("data")
            bg = bg if bg is not None else self.get_data("bg")
            tmp = []
            for wb, dt, sb in zip(w_bkg, data, bg):
                if isinstance(wb, str):
                    wb = self.data.load_weight_file(wb)
                tmp.append(wb * data_shape(dt) / data_shape(sb))
            w_bkg = tmp
            if display:
                print("background weight:", w_bkg)
        else:
            tmp = []
            for wb in w_bkg:
                if isinstance(wb, str):
                    wb = self.data.load_weight_file(wb)
                tmp.append(wb)
            w_bkg = tmp
        return w_bkg, w_inmc

    def get_fcn(self, all_data=None, batch=65000, vm=None, name=""):
        if all_data is None:
            if vm in self.cached_fcn:
                return self.cached_fcn[vm]
            data, phsp, bg, inmc = self.get_all_data()
        else:
            data, phsp, bg, inmc = all_data
        self._Ngroup = len(data)
        if inmc is None:
            inmc = [None] * self._Ngroup
        if bg is None:
            bg = [None] * self._Ngroup
        model = self._get_model(vm=vm, name=name)
        fcns = []
        # print(self.config["data"].get("using_mix_likelihood", False))
        if self.config["data"].get("using_mix_likelihood", False):
            print("  Using Mix Likelihood")
            fcn = MixLogLikehoodFCN(
                model,
                data,
                phsp,
                bg=bg,
                batch=batch,
                gauss_constr=self.gauss_constr_dic,
            )
            if all_data is None:
                self.cached_fcn[vm] = fcn
            return fcn
        for md, dt, mc, sb, ij in zip(model, data, phsp, bg, inmc):
            if self.config["data"].get("model", "auto") == "cfit":
                fcns.append(
                    FCN(
                        md,
                        dt,
                        mc,
                        batch=batch,
                        inmc=ij,
                        gauss_constr=self.gauss_constr_dic,
                    )
                )
            else:
                fcns.append(
                    FCN(
                        md,
                        dt,
                        mc,
                        bg=sb,
                        batch=batch,
                        inmc=ij,
                        gauss_constr=self.gauss_constr_dic,
                    )
                )
        if len(fcns) == 1:
            fcn = fcns[0]
        else:
            fcn = CombineFCN(fcns=fcns, gauss_constr=self.gauss_constr_dic)
        if all_data is None:
            self.cached_fcn[vm] = fcn
        return fcn

    def get_ndf(self):
        amp = self.get_amplitude()
        args_name = amp.vm.trainable_vars
        return len(args_name)

    @staticmethod
    def reweight_init_value(amp, phsp, ns=None):
        """reset decay chain total and make the integration to be ns"""
        total = [i.total for i in amp.decay_group]
        n_phsp = data_shape(phsp)
        weight = np.array(phsp.get("weight", [1] * n_phsp))
        sw = np.sum(weight)
        if ns is None:
            ns = [1] * len(total)
        elif isinstance(ns, (int, float)):
            ns = [ns / len(total)] * len(total)
        for i in total:
            i.set_rho(1.0)
        pw = amp.partial_weight(phsp)
        for i, w, ni in zip(total, pw, ns):
            i.set_rho(np.sqrt(ni / np.sum(weight * w) * sw))

    @time_print
    def fit(
        self,
        data=None,
        phsp=None,
        bg=None,
        inmc=None,
        batch=65000,
        method="BFGS",
        check_grad=False,
        improve=False,
        reweight=False,
        maxiter=None,
    ):
        if data is None and phsp is None:
            data, phsp, bg, inmc = self.get_all_data()
            fcn = self.get_fcn(batch=batch)
        else:
            fcn = self.get_fcn([data, phsp, bg, inmc], batch=batch)
        # print("sss")
        amp = self.get_amplitude()
        print("decay chains included: ")
        for i in self.full_decay:
            ls_list = [getattr(j, "get_ls_list", lambda x: None)() for j in i]
            print("  ", i, " ls: ", *ls_list)
        if reweight:
            ConfigLoader.reweight_init_value(
                amp, phsp[0], ns=data_shape(data[0])
            )

        print("\n########### initial parameters")
        print(json.dumps(amp.get_params(), indent=2), flush=True)
        print("initial NLL: ", fcn({}))  # amp.get_params()))
        # fit configure
        # self.bound_dic[""] = (,)
        self.fit_params = fit(
            fcn=fcn,
            method=method,
            bounds_dict=self.bound_dic,
            check_grad=check_grad,
            improve=False,
            maxiter=maxiter,
        )
        if self.fit_params.hess_inv is not None:
            self.inv_he = self.fit_params.hess_inv
        return self.fit_params

    def reinit_params(self):
        vm = self.get_amplitude().vm
        vm.refresh_vars(init_val=self.init_value, bound_dic=self.bound_dic)

    def fitNtimes(self, N, *args, **kwargs):
        for i in range(N):
            self.reinit_params()
            fit_result = self.fit(*args, **kwargs)
            fit_pars = json.dumps(fit_result.params, indent=2)
            print(fit_pars, flush=True)

    def get_params_error(
        self,
        params=None,
        data=None,
        phsp=None,
        bg=None,
        inmc=None,
        batch=10000,
        using_cached=False,
        method=None,
        force_pos=True,
        correct_params=None,
    ):
        """
        calculate parameters error
        """
        if params is None:
            params = {}
        if correct_params is None:
            correct_params = []
            if method is None:
                method = "correct"
        if data is None:
            data, phsp, bg, inmc = self.get_all_data()
        if hasattr(params, "params"):
            params = getattr(params, "params")
        fcn = self.get_fcn([data, phsp, bg, inmc], batch=batch)
        if using_cached and self.inv_he is not None:
            hesse_error = np.sqrt(np.fabs(self.inv_he.diagonal())).tolist()
        elif method == "3-point":
            self.inv_he = num_hess_inv_3point(fcn, params)
            diag_he = self.inv_he.diagonal()
            hesse_error = np.sqrt(np.fabs(diag_he)).tolist()
        elif method == "correct":
            h = cal_hesse_correct(fcn, params, correct_params)
            if force_pos:
                self.inv_he = force_pos_def(h)
            else:
                self.inv_he = np.linalg.pinv(h)
            diag_he = self.inv_he.diagonal()
            hesse_error = np.sqrt(np.fabs(diag_he)).tolist()
        else:
            hesse_error, self.inv_he = cal_hesse_error(
                fcn,
                params,
                check_posi_def=True,
                save_npy=True,
                force_pos=force_pos,
            )
        # print("parameters order")
        # print(fcn.model.Amp.vm.trainable_vars)
        # print("error matrix:")
        # print(self.inv_he)
        # print("correlation matrix:")
        # print(corr_coef_matrix(self.inv_he))
        print("hesse_error:", hesse_error)
        err = dict(zip(fcn.vm.trainable_vars, hesse_error))
        if hasattr(self, "fit_params"):
            self.fit_params.set_error(err)
        return err

    @classmethod
    def register_function(cls, name=None):
        def _f(f):
            my_name = name
            if my_name is None:
                my_name = f.__name__
            if hasattr(cls, my_name):
                warnings.warn("override function {}".format(name))
            setattr(cls, my_name, f)
            return f

        return _f

    def get_chain(self, idx):
        decay_group = self.full_decay
        return decay_group.get_decay_chain(idx)

    def get_chain_property(self, idx, display=True):
        """Get chain name and curve style in plot"""
        chain = self.get_chain(idx)
        for i in chain:
            curve_style = i.curve_style
            break
        combine = []
        for i in chain:
            if i.core == chain.top:
                combine = list(i.outs)
        names = []
        displays = []
        for i in combine:
            pro = self.particle_property[str(i)]
            names.append(str(i))
            displays.append(pro.get("display", str(i)))
        if display:
            return " ".join(displays), curve_style
        return "_".join(names), curve_style

    def cal_fitfractions(
        self, params={}, mcdata=None, res=None, exclude_res=[], batch=25000
    ):
        if hasattr(params, "params"):
            params = getattr(params, "params")
        if mcdata is None:
            mcdata = self.get_phsp_noeff()
        amp = self.get_amplitude()
        if res is None:
            res = sorted(
                list(set([str(i) for i in amp.res]) - set(exclude_res))
            )
        frac, err_frac = fit_fractions(
            amp, mcdata, self.inv_he, params, batch, res
        )
        return frac, err_frac

    def cal_signal_yields(self, params={}, mcdata=None, batch=25000):
        if hasattr(params, "params"):
            params = getattr(params, "params")
        if mcdata is None:
            mcdata = self.get_data("phsp")
        amp = self.get_amplitude()
        fracs = [
            fit_fractions(amp, i, self.inv_he, params, batch) for i in mcdata
        ]
        data = self.get_data("data")
        bg = self.get_data("bg")
        if bg is None:
            N_total = [data_shape(i) for i in data]
            for i in data:
                N_data = data_shape(i)
                N_total.append((N_data, np.sqrt(N_data)))
        else:
            bg_weight, _ = self._get_bg_weight(data, bg)
            N_total = []
            for i, j, w in zip(data, bg, bg_weight):
                N_data = data_shape(i)
                N_bg = data_shape(j)
                N_total.append(
                    (N_data - w * N_bg, np.sqrt(N_data + w * w * N_bg))
                )

        N_sig_s = []
        for frac_e, N_e in zip(fracs, N_total):
            frac, frac_err = frac_e
            N, N_err = N_e
            N_sig = {}
            for i in frac:
                N_sig[i] = (
                    frac[i] * N,
                    np.sqrt(
                        (N * frac_err.get(i, 0.0)) ** 2
                        + (N_err * frac[i]) ** 2
                    ),
                )
            N_sig_s.append(N_sig)
        return N_sig_s

    def likelihood_profile(self, var, var_min, var_max, N=100):
        params = self.get_params()
        var0 = params[var]
        delta_var = (var_max - var_min) / N
        vm = self.get_amplitude().vm
        unfix = var in vm.get_all_dic(True)
        nlls_up = []
        vars_up = []
        while var0 <= var_max:
            vm.set_fix(var, var0)
            fit_result = self.fit()
            vars_up.append(var0)
            nlls_up.append(fit_result.min_nll)
            var0 += delta_var
        self.set_params(params)
        var0 = params[var] - delta_var
        vars_down = []
        nlls_down = []
        while var0 >= var_min:
            vm.set_fix(var, var0)
            fit_result = self.fit()
            vars_down.append(var0)
            nlls_down.append(fit_result.min_nll)
            var0 -= delta_var
        self.set_params(params)
        vm.set_fix(var, params[var], unfix=unfix)
        return vars_down[::-1] + vars_up, nlls_down[::-1] + nlls_up

    def get_params(self, trainable_only=False):
        return self.get_amplitude().get_params(trainable_only)

    def set_params(self, params, neglect_params=None):
        if isinstance(params, str):
            if params == "":
                return False
            try:
                with open(params) as f:
                    params = yaml.safe_load(f)
            except Exception as e:
                print(e)
                return False
        if hasattr(params, "params"):
            params = params.params
        if isinstance(params, dict):
            if "value" in params:
                params = params["value"]
        amplitude = self.get_amplitude()
        ret = params.copy()
        if neglect_params is None:
            neglect_params = self._neglect_when_set_params
        if len(neglect_params) != 0:
            # warnings.warn("Neglect {} when setting params.".format(neglect_params))
            for v in params:
                if v in self._neglect_when_set_params:
                    del ret[v]
        amplitude.set_params(ret)
        return True

    def save_params(self, file_name):
        params = self.get_params()
        val = {k: float(v) for k, v in params.items()}
        with open(file_name, "w") as f:
            json.dump(val, f, indent=2)

    @contextlib.contextmanager
    def params_trans(self):
        with self.vm.error_trans(self.inv_he) as f:
            yield f

    def save_tensorflow_model(self, dir_name):
        class CustomModule(tf.Module):
            def __init__(self, config_name, share_dict, final_params):
                self.config = ConfigLoader(config_name, share_dict=share_dict)
                self.amp = self.config.get_amplitude()
                self.config.set_params(final_params)
                self.all_variables = self.amp.vm.variables

            @tf.function()
            def __call__(self, *p):
                data = self.config.data.cal_angle(p)
                return self.amp(data)

        module = CustomModule(self.config, self.share_dict, self.get_params())
        n_p = len(self.get_dat_order())
        input_p = [tf.TensorSpec([None, 4], tf.float64) for i in range(n_p)]
        call = module.__call__.get_concrete_function(*input_p)
        tf.saved_model.save(
            module, dir_name, signatures={"serving_default": call}
        )


def set_prefix_constrains(vm, base, params_dic, self):
    prefix = base.get_variable_name()
    p_list = []
    for v in params_dic:
        vname = v
        for tail in ["_range", "_sigma", "_free", "_constr", "_min", "_max"]:
            if v.endswith(tail):
                vname = v[: -len(tail)]
                break

        if vname not in p_list:
            # print(vname, v)
            p_list.append(vname)
            vv = base.get_var(vname)
            # print(vv, prefix + vname)
            # if isinstance(vv, Variable):# getattr(p_i, vname)
            if vv is None:
                continue
            p_sigma = params_dic.get(vname + "_sigma", None)
            if vname in params_dic and params_dic[vname] is not None:
                p_value = params_dic[vname]
                vv.set_value(p_value)
                if p_sigma is None:
                    self.init_value[vname] = p_value
                else:
                    self.init_value[vname] = [p_value, p_sigma]
            else:
                p_value = None
            p_free = params_dic.get(vname + "_free", None)
            if p_free:
                vv.freed()
            elif p_free is False:
                vv.fixed()
            p_range = vname + "_range"
            if p_range in params_dic and params_dic[p_range] is not None:
                lower, upper = params_dic[p_range]
                self.bound_dic[vv.name] = (lower, upper)
                # vm.set_bound({vv.name: (lower, upper)})
            else:
                lower = params_dic.get(vname + "_min")
                upper = params_dic.get(vname + "_max")
                # print(lower, upper)
                if lower is not None or upper is not None:
                    self.bound_dic[vv.name] = (lower, upper)
                    # vm.set_bound({vv.name: (lower, upper)})

                # self.bound_dic[vv.name] = (lower, upper)
            # elif p_sigma is not None and p_value is not None:
            #    p_10sigma = 10 * p_sigma
            #    self.bound_dic[vv.name] = (
            #        p_value - p_10sigma,
            #        p_value + p_10sigma,
            #    )
            p_constr = vname + "_constr"
            if p_constr in params_dic and params_dic[p_constr] is not None:
                if params_dic[p_constr]:
                    if p_value is None:
                        raise Exception(
                            "Need central value of {0} of {1} when adding gaussian constraint".format(
                                vname, prefix
                            )
                        )
                    if p_sigma is None:
                        raise Exception(
                            "Need sigma of {0} of {1} when adding gaussian constraint".format(
                                vname, prefix
                            )
                        )
                    self.gauss_constr_dic[vv.name] = (
                        params_dic[vname],
                        p_sigma,
                    )


def validate_file_name(s):
    rstr = r"[\/\\\:\*\?\"\<\>\|]"  # '/ \ : * ? " < > |'
    name = re.sub(rstr, "_", s)
    return name


class PlotParams(dict):
    def __init__(self, plot_config, decay_struct):
        self.config = plot_config
        self.defaults_config = {}
        self.defaults_config.update(self.config.get("config", {}))
        self.decay_struct = decay_struct
        chain_map = self.decay_struct.get_chains_map()
        self.re_map = {}
        for i in chain_map:
            for _, j in i.items():
                for k, v in j.items():
                    self.re_map[v] = k
        self.params = []
        for i in self.get_mass_vars():
            self.params.append(i)
        for i in self.get_angle_vars():
            self.params.append(i)

    def get_data_index(self, sub, name):
        dec = self.decay_struct.topology_structure()
        if sub == "mass":
            p = get_particle(name)
            return "particle", self.re_map.get(p, p), "m"
        if sub == "p":
            p = get_particle(name)
            return "particle", self.re_map.get(p, p), "p"
        if sub == "angle":
            name_i = name.split("/")
            de_i = self.decay_struct.get_decay_chain(name_i)
            p = get_particle(name_i[-1])
            for i in de_i:
                if p in i.outs:
                    de = i
                    break
            else:
                raise IndexError("not found such decay {}".format(name))
            return (
                "decay",
                de_i.standard_topology(),
                self.re_map.get(de, de),
                self.re_map.get(p, p),
                "ang",
            )
        if sub == "aligned_angle":
            name_i = name.split("/")
            de_i = self.decay_struct.get_decay_chain(name_i)
            p = get_particle(name_i[-1])
            for i in de_i:
                if p in i.outs:
                    de = i
                    break
            else:
                raise IndexError("not found such decay {}".format(name))
            return (
                "decay",
                de_i.standard_topology(),
                self.re_map.get(de, de),
                self.re_map.get(p, p),
                "aligned_angle",
            )
        raise ValueError("unknown sub {}".format(sub))

    def get_mass_vars(self):
        mass = self.config.get("mass", {})
        x = sy.symbols("x")
        for k, v in mass.items():
            display = v.get("display", "M({})".format(k))
            upper_ylim = v.get("upper_ylim", None)
            xrange = v.get("range", None)
            trans = v.get("trans", None)
            if trans is None:
                trans = lambda x: x
            else:
                trans = sy.sympify(trans)
                trans = sy.lambdify(x, trans, modules="numpy")
            units = v.get("units", "GeV")
            bins = v.get("bins", self.defaults_config.get("bins", 50))
            legend = v.get("legend", self.defaults_config.get("legend", True))
            yscale = v.get(
                "yscale", self.defaults_config.get("yscale", "linear")
            )
            yield {
                "name": "m_" + k,
                "display": display,
                "upper_ylim": upper_ylim,
                "idx": (
                    "particle",
                    self.re_map.get(get_particle(k), get_particle(k)),
                    "m",
                ),
                "legend": legend,
                "range": xrange,
                "bins": bins,
                "trans": trans,
                "units": units,
                "yscale": yscale,
            }

    def get_angle_vars(self):
        ang = self.config.get("angle", {})
        for k, i in ang.items():
            names = k.split("/")
            name = names[0]
            number_decay = True
            if len(names) > 1:
                try:
                    count = int(names[-1])
                except ValueError:
                    number_decay = False
            else:
                count = 0
            if number_decay:
                decay_chain, decay = None, None
                part = self.re_map.get(get_particle(name), get_particle(name))
                for decs in self.decay_struct:
                    for dec in decs:
                        if dec.core == get_particle(name):
                            decay = dec.core.decay[count]
                            for j in self.decay_struct:
                                if decay in j:
                                    decay_chain = j.standard_topology()
                            decay = self.re_map.get(decay, decay)
                part = decay.outs[0]
            else:
                _, decay_chain, decay, part, _ = self.get_data_index(
                    "angle", k
                )
            for j, v in i.items():
                display = v.get("display", j)
                upper_ylim = v.get("upper_ylim", None)
                theta = j
                trans = lambda x: x
                if "cos" in j:
                    theta = j[4:-1]
                    trans = np.cos
                bins = v.get("bins", self.defaults_config.get("bins", 50))
                xrange = v.get("range", None)
                legend = v.get(
                    "legend", self.defaults_config.get("legend", False)
                )
                yscale = v.get(
                    "yscale", self.defaults_config.get("yscale", "linear")
                )
                yield {
                    "name": validate_file_name(k + "_" + j),
                    "display": display,
                    "upper_ylim": upper_ylim,
                    "idx": ("decay", decay_chain, decay, part, "ang", theta),
                    "trans": trans,
                    "bins": bins,
                    "range": xrange,
                    "legend": legend,
                    "yscale": yscale,
                }

    def get_params(self, params=None):
        if params is None:
            return self.params
        if isinstance(params, str):
            params = [params]
        params_list = []
        for i in self.params:
            if i["display"] in params:
                params_list.append(i)
        return params_list



import copy

from tf_pwa.amp.core import get_particle_model_name
from tf_pwa.cal_angle import cal_angle_from_momentum
from tf_pwa.config import get_config
from tf_pwa.data import data_mask, data_merge, data_shape
from tf_pwa.particle import BaseParticle
from tf_pwa.phasespace import generate_phsp as generate_phsp_o
from tf_pwa.tensorflow_wrapper import tf

from .config_loader import ConfigLoader


@ConfigLoader.register_function()
def generate_toy(config, N=1000, force=True, max_N=100000):
    decay_group = config.get_decay()
    amp = config.get_amplitude()

    def gen(M):
        return generate_phsp(config, M)

    all_data = []
    n_gen = 0
    n_accept = 0
    n_total = 0
    test_N = 10 * N
    while N > n_accept:
        test_N = abs(min(max_N, test_N))
        data = single_sampling(gen, amp, test_N)
        n_gen = data_shape(data)
        n_total += test_N
        n_accept += n_gen
        test_N = int(1.01 * n_total / (n_accept + 1) * (N - n_accept))
        all_data.append(data)

    ret = data_merge(*all_data)

    if force:
        cut = tf.range(data_shape(ret)) < N
        ret = data_mask(ret, cut)

    return ret


def single_sampling(phsp, amp, N):
    data = phsp(N)
    weight = amp(data)
    rnd = tf.random.uniform(weight.shape, dtype=weight.dtype)
    cut = rnd * tf.reduce_max(weight) * 1.1 < weight
    data = data_mask(data, cut)
    return data


@ConfigLoader.register_function()
def generate_toy2(
    config,
    N=1000,
    force=True,
    gen=None,
    gen_p=None,
    max_N=100000,
    include_charge=False,
):
    """
    A more accurate method for generating toy data.

    :param N: number of events.
    :param force: if romove extra data generated.
    :param gen: optional function for generate phase space, the return value is same as config.get_data.
    :param gen_p:  optional function for generate phase space, the return value is dict as `{B: pb, C: pc, D: pd}`.
    :param max_N: max number of events for every try.

    """

    decay_group = config.get_decay()
    amp = config.get_amplitude()

    if gen is None:
        if gen_p is not None:

            def gen(N):
                p = gen_p(N)
                p = {
                    BaseParticle(k) if isinstance(k, str) else k: v
                    for k, v in p.items()
                }
                return cal_angle_from_momentum(p, config.get_decay(False))

        else:

            def gen(M):
                return generate_phsp(config, M)

    all_data = []
    n_gen = 0
    n_accept = 0
    n_total = 0
    test_N = 10 * N
    if not hasattr(config, "max_amplitude"):
        config.max_amplitude = None

    while N > n_accept:
        test_N = abs(min(max_N, test_N))
        data, new_max_weight = single_sampling2(
            gen,
            amp,
            test_N,
            config.max_amplitude,
            include_charge=include_charge,
        )
        n_gen = data_shape(data)
        n_total += test_N
        if (
            config.max_amplitude is not None
            and new_max_weight > config.max_amplitude
            and len(all_data) > 0
        ):
            tmp = data_merge(*all_data)
            rnd = tf.random.uniform(
                (n_accept,), dtype=config.max_amplitude.dtype
            )
            cut = rnd * new_max_weight / config.max_amplitude < 1.0
            tmp = data_mask(tmp, cut)
            all_data = [tmp]
            n_accept = data_shape(tmp)
        else:
            config.max_amplitude = new_max_weight
        n_accept += n_gen
        test_N = int(1.01 * n_total / (n_accept + 1) * (N - n_accept))
        all_data.append(data)

    ret = data_merge(*all_data)

    if force:
        cut = tf.range(data_shape(ret)) < N
        ret = data_mask(ret, cut)

    return ret


def single_sampling2(phsp, amp, N, max_weight=None, include_charge=False):
    data = phsp(N)
    if "charge_conjugation" not in data:
        if include_charge:
            charge = (
                tf.cast(
                    tf.random.uniform((data_shape(data),)) > 0.5,
                    get_config("dtype"),
                )
                * 2
                - 1
            )
            data["charge_conjugation"] = charge
        else:
            data["charge_conjugation"] = tf.ones(
                (data_shape(data),), get_config("dtype")
            )
    weight = amp(data)
    new_max_weight = tf.reduce_max(weight)
    if max_weight is None or max_weight < new_max_weight:
        max_weight = new_max_weight * 1.01
    rnd = tf.random.uniform(weight.shape, dtype=weight.dtype)
    cut = rnd * max_weight < weight
    data = data_mask(data, cut)
    return data, max_weight


@ConfigLoader.register_function()
def generate_phsp_p(config, N=1000):
    decay_group = config.get_decay()

    m0, mi, idx = build_phsp_chain(decay_group)

    pi = generate_phsp_o(m0, mi, N=N)

    def loop_index(tree, idx):
        for i in idx:
            tree = tree[i]
        return tree

    return {k: loop_index(pi, idx[k]) for k in decay_group.outs}


@ConfigLoader.register_function()
def generate_phsp(config, N=1000):
    p = generate_phsp_p(config, N)
    return cal_angle_from_momentum(p, config.get_decay(False))


def build_phsp_chain(decay_group):
    struct = decay_group.topology_structure()
    inner_node = [set(i.inner) for i in struct]
    a = inner_node[0]
    for i in inner_node[1:]:
        a = a & i

    m0 = decay_group.top.get_mass()
    mi = [i.get_mass() for i in decay_group.outs]

    if any(i is None for i in [m0] + mi):
        raise ValueError("mass required to generate phase space")

    m0 = float(m0)
    mi = [float(i) for i in mi]

    if len(a) == 0:
        return m0, mi, {k: (v,) for v, k in enumerate(decay_group.outs)}

    # print(type(decay_group.get_particle("D")))
    # print([type(i) for i in decay_group[0].inner])

    ref_dec = decay_group[0]

    decay_map = struct[0].topology_map(ref_dec)
    # print(decay_map)
    nodes = []
    for i in a:
        if get_particle_model_name(decay_map[i]) == "one":
            nodes.append((i, float(decay_map[i].get_mass())))

    mi = dict(zip(decay_group.outs, mi))

    st = struct[0].sorted_table()
    mi, final_idx = build_phsp_chain_sorted(st, mi, nodes)
    return m0, mi, final_idx


def build_phsp_chain_sorted(st, final_mi, nodes):
    """
    {A: [B,C, D], R: [B,C]} + {R: M} => ((mr, (mb, mc)), md)
    """
    st = copy.deepcopy(st)
    for i in final_mi:
        del st[i]
    mass_table = final_mi.copy()
    final_idx = {}
    index_root_map = {}
    # print(st)
    max_iter = 10
    while nodes and max_iter > 0:
        pi, mi = nodes.pop(0)
        sub_node = st[pi]
        max_iter -= 1
        if all(i in mass_table for i in sub_node):
            index_root_map[pi] = sub_node
            # the order make the following loop work
            mass_table[pi] = (mi, [mass_table[i] for i in sub_node])
            for k, i in enumerate(sub_node):
                final_idx[i] = (k,)
                del mass_table[i]
            for k, v in st.items():
                if k == pi:
                    continue
                if all(i in v for i in sub_node):
                    for n in sub_node:
                        st[k].remove(n)
                    st[k].append(pi)
        else:
            nodes.append((pi, mi))
    # print(mass_table)
    ret = []
    for k, i in enumerate(mass_table):
        if i in final_mi:
            final_idx[i] = (k,)
        else:
            for j in index_root_map[i]:
                final_idx[j] = (k, *final_idx[j])
        ret.append(mass_table[i])
    # assert False
    return ret, final_idx



import os.path
import sys

import numpy as np

# import tf_pwa
from tf_pwa.applications import gen_data, gen_mc
from tf_pwa.config_loader import ConfigLoader

this_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, this_dir + "/..")


def main():
    """Take three-body decay A->BCD for example, we generate a PhaseSpace MC sample and a toy data sample."""
    import argparse

    parser = argparse.ArgumentParser(
        description="generate toy of a certain amplitude"
    )
    parser.add_argument("--Nmc", default=2000, type=int, dest="Nmc")
    parser.add_argument("--Ndata", default=100, type=int, dest="Ndata")
    results = parser.parse_args()

    if not os.path.exists("data"):
        os.mkdir("data")
    generate_phspMC(Nmc=results.Nmc, mc_file="data/PHSP.dat")
    generate_toy_from_phspMC(
        Ndata=results.Ndata, mc_file="data/PHSP.dat", data_file="data/data.dat"
    )


def generate_phspMC(Nmc, mc_file):
    """Generate PhaseSpace MC of size Nmc and save it as txt file"""
    # the order of A->B,C,D should be same as dat_order in config.yml:data: [B, C, D]
    mA = 4.6  # masses of mother particle A and daughters BCD
    mB = 2.00698
    mC = 2.01028
    mD = 0.13957
    # a2bcd is a [3*Nmc, 4] array, which are the momenta of BCD in the rest frame of A
    a2bcd = gen_mc(mA, [mB, mC, mD], Nmc)
    np.savetxt(mc_file, a2bcd)


def generate_toy_from_phspMC(Ndata, mc_file, data_file):
    """Generate toy using PhaseSpace MC from mc_file"""
    # We use ConfigLoader to read the information in the configuration file
    config = ConfigLoader("config.yml")
    # Set the parameters in the amplitude model
    config.set_params("gen_params.json")
    amp = config.get_amplitude()
    # data is saved in data_file
    data = gen_data(
        amp,
        Ndata=Ndata,
        mcfile=mc_file,  # input phsase space file
        genfile=data_file,  # saved toy data file
        # use the order in config, the default is ascii order.
        particles=config.get_dat_order(),
    )
    return data


if __name__ == "__main__":
    main()



import os.path
import re
import sys

import numpy as np

from tf_pwa.applications import fit_fractions

# import tf_pwa
from tf_pwa.config_loader import ConfigLoader
from tf_pwa.utils import error_print

this_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, this_dir + "/..")


def main():
    """Calculate fit fractions and their errors."""
    import argparse

    parser = argparse.ArgumentParser(
        description="calculate fit fractions and their errors"
    )
    parser.add_argument(
        "--params", default="final_params.json", dest="params_file"
    )
    results = parser.parse_args()

    cal_fitfractions(params_file=results.params_file)


def cal_fitfractions(params_file):
    config = ConfigLoader("config.yml")
    config.set_params(params_file)
    params = config.get_params()
    config.get_params_error(params)

    mcdata = (
        config.get_phsp_noeff()
    )  # use the file of PhaseSpace MC without efficiency indicated in config.yml
    fit_frac, err_frac = fit_fractions(
        config.get_amplitude(), mcdata, config.inv_he, params
    )
    print("########## fit fractions:")
    fit_frac_string = ""
    for i in fit_frac:
        if isinstance(i, tuple):
            name = "{}x{}".format(*i)  # interference term
        else:
            name = i  # fit fraction
        fit_frac_string += "{} {}\n".format(
            name, error_print(fit_frac[i], err_frac.get(i, None))
        )
    print(fit_frac_string)
    print("########## fit fractions table:")
    print_frac_table(
        fit_frac_string
    )  # print the fit-fractions as a 2-D table. The codes below are just to implement the print function.


def print_frac_table(frac_txt):
    def get_point(s):
        partten = re.compile(r"([^\s]+)\s+([+-.e1234567890]+)\s+")
        ret = {}
        for i in s.split("\n"):
            g = partten.match(i)
            if g:
                name = g.group(1).split("x")
                frac = float(g.group(2))
                if len(name) == 1:
                    l, r = name * 2
                elif len(name) == 2:
                    l, r = name
                else:
                    raise Exception("error {}".format(name))
                if l not in ret:
                    ret[l] = {}
                ret[l][r] = frac
        return ret

    s = get_point(frac_txt)

    def get_table(s):
        idx = list(s)
        n_idx = len(idx)
        idx_map = dict(zip(idx, range(n_idx)))
        ret = []
        for i in range(n_idx):
            ret.append([0.0 for j in range(n_idx)])
        for i, k in s.items():
            for j, v in k.items():
                ret[idx_map[i]][idx_map[j]] = v
        return idx, ret

    idx, table = get_table(s)

    ret = []
    for i, k in enumerate(table):
        tmp = []
        for j, v in enumerate(k):
            if i < j:
                tmp.append("-")
            else:
                tmp.append("{:.3f}".format(v))
        ret.append(tmp)
    for i, k in zip(idx, ret):
        print(i, end="\t")
        for v in k:
            print(v, end="\t")
        print()
    print(
        "Total sum:", np.sum(table)
    )  # the sum of all elements in the table, which should be one but for precision
    print(
        "Non-interference sum:", np.sum(np.diagonal(table))
    )  # the sum of all fit-fractions without the interference terms. We expect it to be near one.


if __name__ == "__main__":
    main()



import math

import matplotlib.pyplot as plt

from .particle import split_particle_type


class DotGenerator:
    dot_head = """
digraph {
    rankdir=LR;
    node [shape=point];
    edge [arrowhead=none, labelfloat=true];
"""
    dot_tail = "}\n"
    dot_ranksame = "    {{ rank=same {} }};\n"
    dot_default_node = '    "{}" [shape=none];\n'
    dot_default_edge = '    "{}" -> "{}";\n'
    dot_label_edge = '    "{}" -> "{}" [label="{}"];\n'

    def __init__(self, top):
        self.top = top

    def get_dot_source(self):
        chains = self.top.chain_decay()
        ret = []
        for i in chains:
            dot_source = self.dot_chain(i)
            ret.append(dot_source)
        return ret

    @staticmethod
    def dot_chain(chains, has_label=True):
        ret = DotGenerator.dot_head
        top, _, outs = split_particle_type(chains)

        def format_particle(ps):
            s = ['"{}"'.format(i) for i in ps]
            return ",".join(s)

        for i in top:
            ret += DotGenerator.dot_default_node.format(i)
        for i in outs:
            ret += DotGenerator.dot_default_node.format(i)
        ret += DotGenerator.dot_ranksame.format(format_particle(top))
        ret += DotGenerator.dot_ranksame.format(format_particle(outs))

        decay_dict = {}
        edges = []
        for i in chains:
            if i.core in top:
                edges.append((i.core, i))
            else:
                decay_dict[i.core] = i
            for j in i.outs:
                edges.append((i, j))

        for i, j in edges:
            if j in decay_dict:
                if has_label:
                    ret += DotGenerator.dot_label_edge.format(
                        i, decay_dict[j], j
                    )
                else:
                    ret += DotGenerator.dot_default_edge.format(
                        i, decay_dict[j]
                    )
            else:
                ret += DotGenerator.dot_default_edge.format(i, j)
        ret += DotGenerator.dot_tail
        return ret


def draw_decay_struct(decay_chain, show=False, **kwargs):
    from graphviz import Source

    a = DotGenerator.dot_chain(decay_chain)
    g = Source(a, **kwargs)
    if show:
        g.view()
    else:
        g.render()


def get_node_layout(decay_chain):
    stl = decay_chain.sorted_table_layers()
    decay_map = {}
    max_branchs = max([len(i.outs) for i in decay_chain])
    ys = {}
    xs = {}
    for l, p in enumerate(stl[:0:-1]):
        if p is None:
            continue
        for k, v in p:
            ys[k] = ys.get(k, 0)
            xs[k] = l + 1
            for i in decay_chain:
                if i.core == k:
                    decay_map[k] = i
                    n = len(i.outs)
                    for j, m in enumerate(i.outs):
                        ys[m] = ys[k] + (-j - 0.5 + n / 2) / (
                            max_branchs ** (l + 1) + 2
                        )
                    break

    return xs, ys


def reorder_final_particle(decay_chain, ys):
    stl = decay_chain.sorted_table_layers()
    outs = sorted(decay_chain.outs, key=ys.get)

    ys_new = {k: i for i, k in enumerate(outs)}
    for p in stl[2:]:
        if p is None:
            continue
        for l, (k, v) in enumerate(p):
            ys_new[k] = sum(ys_new.get(i) for i in v) / len(v)
    return ys_new


def get_layout(decay_chain, xs, ys):
    stl = decay_chain.sorted_table_layers()
    points = {"__top": (0, ys[decay_chain.top])}
    for i in xs:
        points[i] = xs[i], ys[i]
    for i in decay_chain.outs:
        points[i] = len(stl) - 1, ys[i]
    lines = [("__top", decay_chain.top)]
    for i in decay_chain:
        for j in i.outs:
            lines.append((i.core, j))
    return lines, points


def get_decay_layout(decay_chain):
    xs, ys = get_node_layout(decay_chain)
    ys = reorder_final_particle(decay_chain, ys)
    lines, points = get_layout(decay_chain, xs, ys)
    return lines, points


def plot_decay_struct(decay_chain, ax=plt):
    lines, points = get_decay_layout(decay_chain)
    for a, b in lines:
        x, y = points[a]
        x2, y2 = points[b]
        ax.arrow(
            x,
            y,
            x2 - x,
            y2 - y,
            width=0.01,
            length_includes_head=True,
            label=str(b),
        )
        rotation = math.atan2(y2 - y, x2 - x) / math.pi * 180
        name = str(b)
        if (x2 - x) * (y2 - y) >= 0:
            ax.text(
                (x + x2) / 2,
                (y + y2) / 2 + 0.01,
                name,
                rotation=rotation,
                ha="center",
                va="bottom",
            )
        else:
            ax.text(
                (x + x2) / 2,
                (y + y2) / 2 + 0.01,
                name,
                rotation=rotation,
                ha="center",
                va="bottom",
            )
    ax.axis("off")



#!/usr/bin/env python

"""
Script for generating data for resolution of gauss function.
"""

import numpy as np
import tensorflow as tf

from tf_pwa.amp import get_particle, get_relative_p2
from tf_pwa.angle import LorentzVector as lv
from tf_pwa.angle import Vector3 as v3
from tf_pwa.cal_angle import cal_helicity_angle
from tf_pwa.config_loader import ConfigLoader


def gauss_random_step(data, decay_chain, r_name, sigma, dat_order):
    angle = cal_helicity_angle(
        data["particle"], decay_chain.standard_topology()
    )
    decay_chain.standard_topology()
    tp_map = decay_chain.topology_map()

    r_particle = tp_map[get_particle(r_name)]

    mass = {}
    for i in data["particle"]:
        mi = data["particle"][i]["m"]
        if i == r_particle:
            mi = mi + tf.random.normal(mi.shape, 0, sigma, dtype=mi.dtype)
        mass[i] = mi

    mask = True
    p4_all = {}
    for i in decay_chain:
        phi = angle[tp_map[i]][tp_map[i.outs[0]]]["ang"]["alpha"]
        theta = angle[tp_map[i]][tp_map[i.outs[0]]]["ang"]["beta"]

        m0 = mass[tp_map[i.core]]
        m1 = mass[tp_map[i.outs[0]]]
        m2 = mass[tp_map[i.outs[1]]]

        mask = mask & (m0 >= m1 + m2)

        p_square = get_relative_p2(m0, m1, m2)

        p = tf.sqrt(tf.where(p_square > 0, p_square, 0))
        pz = p * tf.cos(theta)
        px = p * tf.sin(theta) * tf.cos(phi)
        py = p * tf.sin(theta) * tf.sin(phi)
        E1 = tf.sqrt(m1 * m1 + p * p)
        E2 = tf.sqrt(m2 * m2 + p * p)
        p1 = tf.stack([E1, px, py, pz], axis=-1)
        p2 = tf.stack([E2, -px, -py, -pz], axis=-1)
        p4_all[i.outs[0]] = p1
        p4_all[i.outs[1]] = p2

    core_boost = {}
    for i in decay_chain:
        if i.core != decay_chain.top:
            core_boost[i.outs[0]] = i.core
            core_boost[i.outs[1]] = i.core
    ret = {}
    for i in decay_chain.outs:
        tmp = i
        ret[i] = p4_all[i]
        while tmp in core_boost:
            tmp = core_boost[tmp]
            # print(i, tmp)
            ret[i] = lv.rest_vector(lv.neg(p4_all[tmp]), ret[i])

    ret2 = {}
    mask = tf.expand_dims(mask, -1)
    for i in ret:
        ret2[i] = tf.where(mask, ret[i], data["particle"][tp_map[i]]["p"])
    # print(ret)
    # print({i: data["particle"][tp_map[i]]["p"] for i in decay_chain.outs})

    pi = np.stack([ret2[i] for i in dat_order], axis=1)
    return pi


def gauss_random(data, decay_chain, r_name, sigma, sample_N, dat_order):
    pi_s = []
    for i in range(sample_N):
        pi = gauss_random_step(data, decay_chain, r_name, sigma, dat_order)
        pi_s.append(pi)
    pi = np.concatenate(pi_s, axis=1).reshape((-1, 4))
    return pi


def main():
    sigma = 0.005
    r_name = "R_BC"

    config = ConfigLoader("config.yml")
    sample_N = config.resolution_size

    decays = config.get_decay(False)
    decay_chain = decays.get_decay_chain(r_name)
    data = config.get_data("data_origin")[0]
    phsp = config.get_data("phsp_plot")[0]

    dat_order = config.get_dat_order()

    generate = lambda x: gauss_random(
        x, decay_chain, r_name, sigma, sample_N, dat_order
    )

    pi = generate(data)
    np.savetxt("data/data.dat", pi)
    np.savetxt(
        "data/data_weight.dat", np.ones((pi.shape[0] // len(dat_order),))
    )

    pi = generate(phsp)
    np.save("data/phsp_re.npy", pi)
    np.savetxt(
        "data/phsp_re_weight.dat", np.ones((pi.shape[0] // len(dat_order),))
    )


if __name__ == "__main__":
    main()
